import re
from contextlib import ExitStack, contextmanager
from pathlib import Path
from queue import Queue
from threading import Event
from time import sleep
from unittest import TestCase, skipIf
from unittest.mock import MagicMock, PropertyMock, patch

try:
    import vlc

except (ImportError, OSError):
    vlc = None

from dakara_base.directory import PlatformDirs
from packaging.version import parse

from dakara_player.media_player.base import (
    InvalidStateError,
    KaraFolderNotFound,
    VersionNotFoundError,
    on_playing_this,
)
from dakara_player.media_player.vlc import (
    MediaPlayerVlc,
    VlcTooOldError,
    get_metadata,
    set_metadata,
)
from dakara_player.mrl import path_to_mrl
from dakara_player.text import TextGenerator
from dakara_player.window import DummyWindowManager, WindowManager
from tests.utils import get_temp_dir


class BaseTestCase(TestCase):
    """Base test case to provid a mocked instance of MediaPlayerVlc."""

    def setUp(self):
        # create playlist entry ID
        self.id = 42

        # create playlist entry file path
        self.song_file_path = Path("file")

        # create playlist entry
        self.playlist_entry = {
            "id": self.id,
            "song": {"title": "Song title", "file_path": self.song_file_path},
            "owner": "me",
            "use_instrumental": False,
        }

    @contextmanager
    def get_instance(
        self,
        config=None,
        tempdir=None,
    ):
        """Get a heavily mocked instance of MediaPlayerVlc.

        Args:
            config (dict): Configuration passed to the constructor.
            tempdir (pathlib.Path): Path to temporary directory.

        Yields:
            tuple: Contains the following elements:
                MediaPlayerVlc: Instance;
                tuple: Contains the mocked objects:
                    unittest.mock.MagicMock: VLC Instance object.
                    unittest.mock.MagicMock: BackgroundLoader object.
                    unittest.mock.MagicMock: TextGenerator object.
                tuple: Contains the mocked classes:
                    unittest.mock.MagicMock: VLC Instance class.
                    unittest.mock.MagicMock: BackgroundLoader class.
                    unittest.mock.MagicMock: TextGenerator class.
        """
        config = config or {"kara_folder": str(get_temp_dir())}

        with ExitStack() as stack:
            if vlc is None:
                mocked_vlc = stack.enter_context(
                    patch("dakara_player.media_player.vlc.vlc")
                )
                mocked_instance_class = mocked_vlc.Instance
            else:
                mocked_instance_class = stack.enter_context(
                    patch("dakara_player.media_player.vlc.vlc.Instance")
                )

            mocked_background_loader_class = stack.enter_context(
                patch("dakara_player.media_player.base.BackgroundLoader")
            )

            mocked_text_generator_class = stack.enter_context(
                patch("dakara_player.media_player.base.TextGenerator")
            )

            stack.enter_context(patch.object(MediaPlayerVlc, "check_is_available"))

            if tempdir is None:
                tempdir = Path("temp")

            yield (
                MediaPlayerVlc(Event(), Queue(), config, tempdir),
                (
                    mocked_instance_class.return_value,
                    mocked_background_loader_class.return_value,
                    mocked_text_generator_class.return_value,
                ),
                (
                    mocked_instance_class,
                    mocked_background_loader_class,
                    mocked_text_generator_class,
                ),
            )


@patch("dakara_player.media_player.base.TRANSITION_DURATION", 10)
@patch("dakara_player.media_player.base.IDLE_DURATION", 20)
class MediaPlayerVlcTestCase(BaseTestCase):
    """Test the VLC player class unitary."""

    def set_playlist_entry(self, vlc_player, started=True):
        """Set a playlist entry and make the player play it.

        Args:
            vlc_player (MediaPlayerVlc): Instance of the VLC player.
            started (bool): If True, make the player play the song.
        """
        vlc_player.playlist_entry = self.playlist_entry

        # create mocked transition
        vlc_player.playlist_entry_data["transition"].media = MagicMock()

        # create mocked song
        media_song = MagicMock()
        media_song.get_mrl.return_value = path_to_mrl(
            vlc_player.kara_folder_path / self.playlist_entry["song"]["file_path"]
        )
        vlc_player.playlist_entry_data["song"].media = media_song

        # set media has started
        if started:
            player = vlc_player.instance.media_player_new.return_value
            player.get_media.return_value = vlc_player.playlist_entry_data["song"].media
            vlc_player.playlist_entry_data["transition"].started = True
            vlc_player.playlist_entry_data["song"].started = True

    def test_init_window(self):
        """Test to use default or custom window."""
        # default window
        with self.get_instance(
            {"kara_folder": str(get_temp_dir()), "vlc": {"use_default_window": True}}
        ) as (vlc_player, _, _):
            self.assertIsInstance(vlc_player.window, DummyWindowManager)

        # custom window
        with self.get_instance() as (vlc_player, _, _):
            self.assertIsInstance(vlc_player.window, WindowManager)

    def test_set_callback(self):
        """Test the assignation of a callback."""
        with self.get_instance() as (vlc_player, _, _):
            # create a callback function
            callback = MagicMock()

            # pre assert the callback is not set yet
            self.assertIsNot(vlc_player.callbacks.get("test"), callback)

            # call the method
            vlc_player.set_callback("test", callback)

            # post assert the callback is now set
            self.assertIs(vlc_player.callbacks.get("test"), callback)

    @skipIf(vlc is None, "VLC not installed")
    def test_set_vlc_callback(self):
        """Test the assignation of a callback to a VLC event.

        We have also to mock the event manager method because there is no way
        with the VLC library to know which callback is associated to a given
        event.
        """
        with self.get_instance() as (vlc_player, _, _):
            # patch the event creator
            vlc_player.event_manager.event_attach = MagicMock()

            # create a callback function
            callback = MagicMock()

            # pre assert the callback is not set yet
            self.assertIsNot(
                vlc_player.vlc_callbacks.get(vlc.EventType.MediaPlayerEndReached),
                callback,
            )

            # call the method
            vlc_player.set_vlc_callback(vlc.EventType.MediaPlayerEndReached, callback)

            # assert the callback is now set
            self.assertIs(
                vlc_player.vlc_callbacks.get(vlc.EventType.MediaPlayerEndReached),
                callback,
            )

            # assert the event manager got the right arguments
            vlc_player.event_manager.event_attach.assert_called_with(
                vlc.EventType.MediaPlayerEndReached, callback
            )

    @skipIf(vlc is None, "VLC not installed")
    def test_vlc_unavailable(self):
        """Test that is_available returns False when vlc.Instance raises a NameError."""
        with patch.object(vlc, "Instance", side_effect=NameError()):
            self.assertFalse(MediaPlayerVlc.is_available())

    @patch("dakara_player.media_player.vlc.libvlc_get_version")
    def test_get_version_long_4_digits(self, mocked_libvlc_get_version):
        """Test to get the VLC version when it is long and contains 4 digits."""
        # mock the version of VLC
        mocked_libvlc_get_version.return_value = b"3.0.11.1 Vetinari"

        # call the method
        version = MediaPlayerVlc.get_version()

        # assert the result
        self.assertEqual(version, parse("3.0.11.1"))

    @patch("dakara_player.media_player.vlc.libvlc_get_version")
    def test_get_version_long(self, mocked_libvlc_get_version):
        """Test to get the VLC version when it is long."""
        # mock the version of VLC
        mocked_libvlc_get_version.return_value = b"3.0.11 Vetinari"

        # call the method
        version = MediaPlayerVlc.get_version()

        # assert the result
        self.assertEqual(version, parse("3.0.11"))

    @patch("dakara_player.media_player.vlc.libvlc_get_version")
    def test_get_version_not_found(self, mocked_libvlc_get_version):
        """Test to get the VLC version when it is not available."""
        # mock the version of VLC
        mocked_libvlc_get_version.return_value = b"none"

        # call the method
        with self.assertRaisesRegex(VersionNotFoundError, "Unable to get VLC version"):
            MediaPlayerVlc.get_version()

    @patch.object(MediaPlayerVlc, "get_version")
    def test_get_version_str(self, mocked_get_version):
        """Test to get the VLC version as a string."""
        # mock the version of VLC
        mocked_get_version.return_value = parse("3.0.11")

        # call the method
        version_str = MediaPlayerVlc.get_version_str()

        # assert the result
        self.assertEqual(version_str, "3.0.11")

    @patch.object(MediaPlayerVlc, "get_version")
    def test_get_version_str_not_found(self, mocked_get_version):
        """Test to get the VLC version as a string when it is not
        available."""
        # mock the version of VLC
        mocked_get_version.side_effect = VersionNotFoundError(
            "Unable to get VLC version"
        )

        # call the method
        self.assertEqual(MediaPlayerVlc.get_version_str(), "unknown version")

    @patch.object(MediaPlayerVlc, "get_version")
    def test_check_version(self, mocked_get_version):
        """Test to check recent enough version VLC."""
        with self.get_instance() as (vlc_player, _, _):
            # mock the version of VLC
            mocked_get_version.return_value = parse("3.0.0")

            # call the method
            vlc_player.check_version()

    @patch.object(MediaPlayerVlc, "get_version")
    def test_check_version_old(self, mocked_get_version):
        """Test to check old version of VLC."""
        with self.get_instance() as (vlc_player, _, _):
            # mock the version of VLC
            mocked_get_version.return_value = parse("2.0.0")

            # call the method
            with self.assertRaisesRegex(VlcTooOldError, "VLC is too old"):
                vlc_player.check_version()

    @patch.object(MediaPlayerVlc, "get_version")
    def test_check_version_problem(self, mocked_get_version):
        """Test to check known problematic version of VLC."""
        with self.get_instance() as (vlc_player, _, _):
            versions = ["3.0.13", "3.0.14", "3.0.15", "3.0.16"]

            for version in versions:
                # mock the version of VLC
                mocked_get_version.return_value = parse(version)

                # call the method
                with self.assertLogs(
                    "dakara_player.media_player.vlc", "WARNING"
                ) as logger:
                    vlc_player.check_version()

                self.assertListEqual(
                    logger.output,
                    [
                        "WARNING:dakara_player.media_player.vlc:This version of VLC "
                        "is known to not work with Dakara player"
                    ],
                )

    @patch.object(Path, "exists")
    def test_check_kara_folder_path(self, mocked_exists):
        """Test to check if the kara folder exists."""
        with self.get_instance() as (vlc_player, _, _):
            # pretend the directory exists
            mocked_exists.return_value = True

            # call the method
            vlc_player.check_kara_folder_path()

            # assert the call
            mocked_exists.assert_called_with()

    @patch.object(Path, "exists")
    def test_check_kara_folder_path_does_not_exist(self, mocked_exists):
        """Test to check if the kara folder does not exist."""
        with self.get_instance() as (vlc_player, _, _):
            # pretend the directory does not exist
            mocked_exists.return_value = False

            # call the method
            with self.assertRaisesRegex(
                KaraFolderNotFound,
                'Karaoke folder "{}" does not exist'.format(
                    re.escape(str(get_temp_dir()))
                ),
            ):
                vlc_player.check_kara_folder_path()

    @patch.object(WindowManager, "get_id")
    @patch.object(WindowManager, "open")
    @patch.object(MediaPlayerVlc, "check_kara_folder_path")
    @patch.object(MediaPlayerVlc, "check_version")
    @patch.object(MediaPlayerVlc, "set_vlc_default_callbacks")
    @patch.object(MediaPlayerVlc, "get_version")
    def test_load(
        self,
        mocked_get_version,
        mocked_set_vlc_default_callback,
        mocked_check_version,
        mocked_check_kara_folder_path,
        mocked_open,
        mocked_get_id,
    ):
        """Test to load the instance."""
        with self.get_instance() as (
            vlc_player,
            (_, mocked_background_loader, mocked_text_generator),
            _,
        ):
            # setup mocks
            mocked_get_version.return_value = "3.0.0 NoName"

            # call the method
            with self.assertLogs("dakara_player.media_player.vlc", "INFO") as logger:
                vlc_player.load()

            # assert the calls
            mocked_check_kara_folder_path.assert_called_with()
            mocked_text_generator.load.assert_called_with()
            mocked_background_loader.load.assert_called_with()
            mocked_check_version.assert_called_with()
            mocked_set_vlc_default_callback.assert_called_with()
            mocked_open.assert_called_with()
            mocked_get_id.assert_called_with()

            # assert logs
            self.assertListEqual(
                logger.output, ["INFO:dakara_player.media_player.vlc:VLC 3.0.0 NoName"]
            )

    @patch.object(Path, "is_file")
    def test_set_playlist_entry_error_file(self, mocked_is_file):
        """Test to set a playlist entry that does not exist."""
        with self.get_instance() as (vlc_player, _, _):
            # mock the system call
            mocked_is_file.return_value = False

            # mock the callbacks
            vlc_player.set_callback("could_not_play", MagicMock())
            vlc_player.set_callback("error", MagicMock())

            # pre assertions
            self.assertIsNone(vlc_player.playlist_entry)

            # call the method
            with self.assertLogs("dakara_player.media_player.base", "DEBUG") as logger:
                vlc_player.set_playlist_entry(self.playlist_entry)

            # call assertions
            mocked_is_file.assert_called_once_with()

            # post assertions
            self.assertIsNone(vlc_player.playlist_entry)

            # assert the callbacks
            vlc_player.callbacks["could_not_play"].assert_called_with(self.id)
            vlc_player.callbacks["error"].assert_called_with(self.id, "File not found")

            # assert the effects on logs
            self.assertListEqual(
                logger.output,
                [
                    "ERROR:dakara_player.media_player.base:File not found '{}'".format(
                        get_temp_dir() / self.song_file_path
                    )
                ],
            )

    @patch("dakara_player.media_player.vlc.set_metadata")
    @patch("dakara_player.media_player.vlc.get_metadata")
    @patch.object(MediaPlayerVlc, "manage_instrumental")
    @patch.object(MediaPlayerVlc, "play")
    @patch.object(MediaPlayerVlc, "generate_text")
    @patch.object(Path, "is_file")
    def test_set_playlist_entry(
        self,
        mocked_is_file,
        mocked_generate_text,
        mocked_play,
        mocked_manage_instrumental,
        mocked_get_metadata,
        mocked_set_metadata,
    ):
        """Test to set a playlist entry."""
        with self.get_instance() as (vlc_player, (_, mocked_background_loader, _), _):
            # setup mocks
            mocked_is_file.return_value = True
            mocked_background_loader.backgrounds = {
                "transition": get_temp_dir() / "transition.png"
            }

            # mock the callbacks
            vlc_player.set_callback("could_not_play", MagicMock())
            vlc_player.set_callback("error", MagicMock())

            # pre assertions
            self.assertIsNone(vlc_player.playlist_entry)

            # call the method
            vlc_player.set_playlist_entry(self.playlist_entry)
            self.assertFalse(self.playlist_entry["use_instrumental"])

            # post assertions
            self.assertDictEqual(vlc_player.playlist_entry, self.playlist_entry)

            # assert the callbacks
            vlc_player.callbacks["could_not_play"].assert_not_called()
            vlc_player.callbacks["error"].assert_not_called()

            # assert mocks
            mocked_is_file.assert_called_with()
            mocked_generate_text.assert_called_with("transition")
            mocked_play.assert_called_with("transition")
            mocked_manage_instrumental.assert_not_called()

    @patch.object(MediaPlayerVlc, "get_track_id_audio_list")
    @patch.object(MediaPlayerVlc, "get_number_tracks")
    @patch.object(MediaPlayerVlc, "get_instrumental_file")
    def test_manage_instrumental_file(
        self,
        mocked_get_instrumental_file,
        mocked_get_number_tracks,
        mocked_get_track_id_audio_list,
    ):
        """Test to add instrumental file."""
        with self.get_instance() as (vlc_player, (mocked_instance, _, _), _):
            video_path = get_temp_dir() / "video"
            audio_path = get_temp_dir() / "audio"

            # pre assertions
            self.assertIsNone(vlc_player.playlist_entry_data["song"].track_id_audio)
            self.assertIsNotNone(vlc_player.kara_folder_path)

            # set playlist entry to request instrumental
            self.playlist_entry["use_instrumental"] = True

            # mocks
            mocked_get_instrumental_file.return_value = audio_path
            mocked_get_number_tracks.return_value = 2
            mocked_media_song = mocked_instance.media_new_path.return_value
            vlc_player.playlist_entry_data["song"].media = mocked_media_song

            # call the method
            with self.assertLogs("dakara_player.media_player.vlc", "DEBUG") as logger:
                vlc_player.manage_instrumental(self.playlist_entry, video_path)

            # post assertions
            self.assertEqual(vlc_player.playlist_entry_data["song"].track_id_audio, 2)

            # assert the effects on logs
            self.assertListEqual(
                logger.output,
                [
                    "INFO:dakara_player.media_player.vlc:Requesting to play "
                    "instrumental file '{}' for '{}'".format(audio_path, video_path),
                ],
            )

            # assert the call
            mocked_get_track_id_audio_list.assert_not_called()

    @patch.object(MediaPlayerVlc, "get_number_tracks")
    @patch.object(MediaPlayerVlc, "get_instrumental_file")
    def test_manage_instrumental_file_error_slaves_add(
        self,
        mocked_get_instrumental_file,
        mocked_get_number_tracks,
    ):
        """Test to be unable to add instrumental file."""
        with self.get_instance() as (vlc_player, (mocked_instance, _, _), _):
            video_path = get_temp_dir() / "video"
            audio_path = get_temp_dir() / "audio"

            # pre assertions
            self.assertIsNone(vlc_player.playlist_entry_data["song"].track_id_audio)
            self.assertIsNotNone(vlc_player.kara_folder_path)

            # set playlist entry to request instrumental
            self.playlist_entry["use_instrumental"] = True

            # mocks
            mocked_get_instrumental_file.return_value = audio_path
            mocked_get_number_tracks.return_value = 2

            # make slaves_add method unavailable
            mocked_media_song = mocked_instance.return_value.media_new_path.return_value
            mocked_media_song.slaves_add.side_effect = NameError("no slaves_add")
            vlc_player.playlist_entry_data["song"].media = mocked_media_song

            # call the method
            with self.assertLogs("dakara_player.media_player.vlc", "DEBUG") as logger:
                vlc_player.manage_instrumental(self.playlist_entry, video_path)

            # post assertions
            self.assertIsNone(vlc_player.playlist_entry_data["song"].track_id_audio)

            # assert the effects on logs
            self.assertListEqual(
                logger.output,
                [
                    "INFO:dakara_player.media_player.vlc:Requesting to play "
                    "instrumental file '{}' for '{}'".format(audio_path, video_path),
                    "ERROR:dakara_player.media_player.vlc:This version of VLC does "
                    "not support slaves, cannot add instrumental file",
                ],
            )

    @patch.object(MediaPlayerVlc, "get_track_id_audio_list")
    @patch.object(MediaPlayerVlc, "get_number_tracks")
    @patch.object(MediaPlayerVlc, "get_instrumental_file")
    def test_manage_instrumental_track(
        self,
        mocked_get_instrumental_file,
        mocked_get_number_tracks,
        mocked_get_track_id_audio_list,
    ):
        """Test add instrumental track."""
        with self.get_instance() as (
            vlc_player,
            (
                mocked_instance,
                _,
                _,
            ),
            _,
        ):
            video_path = get_temp_dir() / "video"

            # pre assertions
            self.assertIsNone(vlc_player.playlist_entry_data["song"].track_id_audio)
            self.assertIsNotNone(vlc_player.kara_folder_path)

            # set playlist entry to request instrumental
            self.playlist_entry["use_instrumental"] = True

            # mocks
            mocked_get_instrumental_file.return_value = None
            mocked_get_track_id_audio_list.return_value = [0, 99, 42]
            mocked_media_song = mocked_instance.media_new_path.return_value
            vlc_player.playlist_entry_data["song"].media = mocked_media_song

            # call the method
            with self.assertLogs("dakara_player.media_player.vlc", "DEBUG") as logger:
                vlc_player.manage_instrumental(self.playlist_entry, video_path)

            # post assertions
            self.assertEqual(vlc_player.playlist_entry_data["song"].track_id_audio, 99)

            # assert the effects on logs
            self.assertListEqual(
                logger.output,
                [
                    "INFO:dakara_player.media_player.vlc:Requesting to play "
                    "instrumental track of '{}'".format(video_path),
                ],
            )

            # assert the call
            mocked_get_number_tracks.assert_not_called()

    @patch.object(MediaPlayerVlc, "get_track_id_audio_list")
    @patch.object(MediaPlayerVlc, "get_instrumental_file")
    def test_manage_instrumental_no_instrumental_found(
        self, mocked_get_instrumental_file, mocked_get_track_id_audio_list
    ):
        """Test to cannot find instrumental."""
        with self.get_instance() as (vlc_player, (mocked_instance, _, _), _):
            video_path = get_temp_dir() / "video"

            # pre assertions
            self.assertIsNone(vlc_player.playlist_entry_data["song"].track_id_audio)

            # set playlist entry to request instrumental
            self.playlist_entry["use_instrumental"] = True

            # mocks
            mocked_get_instrumental_file.return_value = None
            mocked_get_track_id_audio_list.return_value = [99]

            # make slaves_add method unavailable
            mocked_media_song = mocked_instance.return_value.media_new_path.return_value
            vlc_player.playlist_entry_data["song"].media = mocked_media_song

            # call the method
            with self.assertLogs("dakara_player.media_player.vlc", "DEBUG") as logger:
                vlc_player.manage_instrumental(self.playlist_entry, video_path)

            # post assertions
            self.assertIsNone(vlc_player.playlist_entry_data["song"].track_id_audio)

            # assert the effects on logs
            self.assertListEqual(
                logger.output,
                [
                    "WARNING:dakara_player.media_player.vlc:Cannot find instrumental "
                    "file or track for file '{}'".format(video_path)
                ],
            )

    @patch.object(MediaPlayerVlc, "is_playing_this")
    def test_set_pause_idle(self, mocked_is_playing_this):
        """Test to set pause when the player is idle."""
        with self.get_instance() as (vlc_player, (mocked_instance, _, _), _):
            player = mocked_instance.media_player_new.return_value

            # mock
            mocked_is_playing_this.side_effect = lambda what: what == "idle"

            # call method
            vlc_player.pause()

            # assert call
            player.pause.assert_not_called()

    @patch.object(MediaPlayerVlc, "is_playing_this")
    def test_restart_transition(self, mocked_is_playing_this):
        """Test to restart on transition screen."""
        with self.get_instance() as (vlc_player, (mocked_instance, _, _), _):
            player = mocked_instance.media_player_new.return_value

            # mock
            mocked_is_playing_this.side_effect = lambda what: what == "transition"

            # call method
            vlc_player.restart()

            # assert call
            player.set_time.assert_not_called()

    @patch.object(MediaPlayerVlc, "is_playing_this")
    @patch.object(MediaPlayerVlc, "clear_playlist_entry")
    def test_skip_idle(self, mocked_clear_playlist_entry, mocked_is_playing_this):
        """Test to skip on idle screen."""
        with self.get_instance() as (vlc_player, (mocked_instance, _, _), _):
            # mock
            mocked_is_playing_this.side_effect = lambda what: what == "idle"

            # call method
            vlc_player.skip(True)

            # assert call
            vlc_player.clear_playlist_entry.assert_not_called()

    @patch.object(MediaPlayerVlc, "is_playing_this")
    def test_rewind_transition(self, mocked_is_playing_this):
        """Test to rewind on transition screen."""
        with self.get_instance() as (vlc_player, (mocked_instance, _, _), _):
            player = mocked_instance.media_player_new.return_value

            # mock
            mocked_is_playing_this.side_effect = lambda what: what == "transition"

            # call method
            vlc_player.rewind()

            # assert call
            player.set_time.assert_not_called()

    @patch.object(MediaPlayerVlc, "is_playing_this")
    def test_fast_forward_transition(self, mocked_is_playing_this):
        """Test to advance on transition screen."""
        with self.get_instance() as (vlc_player, (mocked_instance, _, _), _):
            player = mocked_instance.media_player_new.return_value

            # mock
            mocked_is_playing_this.side_effect = lambda what: what == "transition"

            # call method
            vlc_player.fast_forward()

            # assert call
            player.set_time.assert_not_called()

    @patch.object(MediaPlayerVlc, "create_thread")
    @patch.object(MediaPlayerVlc, "is_playing_this")
    def test_handle_end_reached_transition(
        self, mocked_is_playing_this, mocked_create_thread
    ):
        """Test song end callback after a transition screen."""
        with self.get_instance() as (vlc_player, _, _):
            self.set_playlist_entry(vlc_player)
            vlc_player.playlist_entry_data["song"].started = False

            # mock the call
            mocked_is_playing_this.side_effect = lambda what: what == "transition"
            vlc_player.set_callback("finished", MagicMock())

            # call the method
            with self.assertLogs("dakara_player.media_player.vlc", "DEBUG") as logger:
                vlc_player.handle_end_reached("event")

            # assert effect on logs
            self.assertListEqual(
                logger.output,
                [
                    "DEBUG:dakara_player.media_player.vlc:End reached callback called",
                    "DEBUG:dakara_player.media_player.vlc:Will play '{}'".format(
                        get_temp_dir() / self.song_file_path
                    ),
                ],
            )

            # assert the call
            vlc_player.callbacks["finished"].assert_not_called()
            mocked_create_thread.assert_called_with(
                target=vlc_player.play, args=("song",)
            )

    @patch.object(MediaPlayerVlc, "create_thread")
    @patch.object(MediaPlayerVlc, "is_playing_this")
    def test_handle_end_reached_song(
        self, mocked_is_playing_this, mocked_create_thread
    ):
        """Test song end callback after a song."""
        with self.get_instance() as (vlc_player, _, _):
            self.set_playlist_entry(vlc_player)

            # mock the call
            vlc_player.set_callback("finished", MagicMock())
            mocked_is_playing_this.side_effect = lambda what: what == "song"

            # call the method
            with self.assertLogs("dakara_player.media_player.vlc", "DEBUG"):
                vlc_player.handle_end_reached("event")

            # post assert
            self.assertIsNone(vlc_player.playlist_entry_data["song"].media)

            # assert the call
            vlc_player.callbacks["finished"].assert_called_with(42)
            mocked_create_thread.assert_not_called()

    @patch.object(MediaPlayerVlc, "create_thread")
    @patch.object(MediaPlayerVlc, "is_playing_this")
    def test_handle_end_reached_song_too_fast(
        self, mocked_is_playing_this, mocked_create_thread
    ):
        """Test song end callback after a song with instantaneous response
        from server.
        """
        with self.get_instance() as (vlc_player, _, _):
            self.set_playlist_entry(vlc_player)

            # mock the call
            vlc_player.set_callback(
                "finished", lambda _: self.set_playlist_entry(vlc_player, started=False)
            )
            mocked_is_playing_this.side_effect = lambda what: what == "song"

            # call the method
            with self.assertLogs("dakara_player.media_player.vlc", "DEBUG"):
                vlc_player.handle_end_reached("event")

            # post assert
            self.assertTrue(vlc_player.errors.empty())
            self.assertIsNotNone(vlc_player.playlist_entry_data["song"].media)

            # assert the call
            mocked_create_thread.assert_not_called()

    @patch.object(MediaPlayerVlc, "create_thread")
    @patch.object(MediaPlayerVlc, "is_playing_this")
    def test_handle_end_reached_idle(
        self, mocked_is_playing_this, mocked_create_thread
    ):
        """Test song end callback after an idle screen."""
        with self.get_instance() as (vlc_player, _, _):
            self.set_playlist_entry(vlc_player)

            # mock the call
            vlc_player.set_callback("finished", MagicMock())
            mocked_is_playing_this.side_effect = lambda what: what == "idle"

            # call the method
            with self.assertLogs("dakara_player.media_player.vlc", "DEBUG"):
                vlc_player.handle_end_reached("event")

            # assert the call
            vlc_player.callbacks["finished"].assert_not_called()
            mocked_create_thread.assert_called_with(
                target=vlc_player.play, args=("idle",)
            )

    @patch.object(MediaPlayerVlc, "create_thread")
    @patch.object(MediaPlayerVlc, "is_playing_this")
    def test_handle_end_reached_invalid(
        self, mocked_is_playing_this, mocked_create_thread
    ):
        """Test song end callback on invalid state."""
        with self.get_instance() as (vlc_player, _, _):
            self.set_playlist_entry(vlc_player)

            # mock the call
            vlc_player.set_callback("finished", MagicMock())
            mocked_is_playing_this.return_value = False

            self.assertFalse(vlc_player.stop.is_set())

            # call the method
            with self.assertLogs("dakara_player.media_player.vlc", "DEBUG"):
                vlc_player.handle_end_reached("event")

            self.assertTrue(vlc_player.stop.is_set())
            exception_class, _, _ = vlc_player.errors.get()
            self.assertIs(InvalidStateError, exception_class)

            # assert the call
            vlc_player.callbacks["finished"].assert_not_called()
            mocked_create_thread.assert_not_called()

    @patch.object(MediaPlayerVlc, "skip")
    @patch.object(MediaPlayerVlc, "is_playing_this")
    def test_handle_encountered_error(self, mocked_is_playing_this, mocked_skip):
        """Test error callback."""
        with self.get_instance() as (vlc_player, _, _):
            self.set_playlist_entry(vlc_player)

            # mock the call
            mocked_is_playing_this.side_effect = lambda what: what == "song"
            vlc_player.set_callback("error", MagicMock())

            # call the method
            with self.assertLogs("dakara_player.media_player.vlc", "DEBUG") as logger:
                vlc_player.handle_encountered_error("event")

            # assert effect on logs
            self.assertListEqual(
                logger.output,
                [
                    "DEBUG:dakara_player.media_player.vlc:Error callback called",
                    "ERROR:dakara_player.media_player.vlc:Unable to play '{}'".format(
                        get_temp_dir() / self.song_file_path
                    ),
                ],
            )

            # assert the call
            vlc_player.callbacks["error"].assert_called_with(
                42, "Unable to play current song"
            )
            mocked_skip.assert_called_with()

    @patch.object(MediaPlayerVlc, "get_timing")
    @patch.object(MediaPlayerVlc, "is_playing_this")
    def test_handle_playing_resumed(self, mocked_is_playing_this, mocked_get_timing):
        """Test playing callback when resuming."""
        with self.get_instance() as (vlc_player, _, _):
            self.set_playlist_entry(vlc_player)

            # mock the call
            vlc_player.set_callback("resumed", MagicMock())
            mocked_is_playing_this.side_effect = lambda what: what == "song"
            mocked_get_timing.return_value = 25

            # call the method
            with self.assertLogs("dakara_player.media_player.vlc", "DEBUG") as logger:
                vlc_player.handle_playing("event")

            # assert effect on logs
            self.assertListEqual(
                logger.output,
                [
                    "DEBUG:dakara_player.media_player.vlc:Playing callback called",
                    "DEBUG:dakara_player.media_player.vlc:Resumed play",
                ],
            )

            # assert the call
            vlc_player.callbacks["resumed"].assert_called_with(42, 25)
            mocked_is_playing_this.assert_called_with("song")

    @patch.object(MediaPlayerVlc, "is_playing_this")
    def test_handle_playing_transition_starts(self, mocked_is_playing_this):
        """Test playing callback when transition starts."""
        with self.get_instance() as (vlc_player, _, _):
            self.set_playlist_entry(vlc_player, started=False)

            # mock the call
            vlc_player.set_callback("started_transition", MagicMock())
            vlc_player.playlist_entry_data["transition"].started = False
            mocked_is_playing_this.side_effect = lambda what: what == "transition"

            # pre assert
            self.assertFalse(vlc_player.playlist_entry_data["transition"].started)
            self.assertFalse(vlc_player.playlist_entry_data["song"].started)

            # call the method
            with self.assertLogs("dakara_player.media_player.vlc", "DEBUG") as logger:
                vlc_player.handle_playing("event")

            # assert effect on logs
            self.assertListEqual(
                logger.output,
                [
                    "DEBUG:dakara_player.media_player.vlc:Playing callback called",
                    "INFO:dakara_player.media_player.vlc:Playing transition for "
                    "'Song title'",
                ],
            )

            # post assert
            self.assertTrue(vlc_player.playlist_entry_data["transition"].started)
            self.assertFalse(vlc_player.playlist_entry_data["song"].started)

            # assert the call
            vlc_player.callbacks["started_transition"].assert_called_with(42)

    @patch.object(MediaPlayerVlc, "is_playing_this")
    def test_handle_playing_song(self, mocked_is_playing_this):
        """Test playing callback when song starts."""
        with self.get_instance() as (vlc_player, _, _):
            self.set_playlist_entry(vlc_player)
            vlc_player.playlist_entry_data["song"].started = False

            # mock the call
            vlc_player.set_callback("started_song", MagicMock())
            vlc_player.playlist_entry_data["transition"].started = True
            vlc_player.playlist_entry_data["song"].started = False
            mocked_is_playing_this.side_effect = lambda what: what == "song"

            # pre assert
            self.assertFalse(vlc_player.playlist_entry_data["song"].started)

            # call the method
            with self.assertLogs("dakara_player.media_player.vlc", "DEBUG") as logger:
                vlc_player.handle_playing("event")

            # assert effect on logs
            self.assertListEqual(
                logger.output,
                [
                    "DEBUG:dakara_player.media_player.vlc:Playing callback called",
                    "INFO:dakara_player.media_player.vlc:Now playing 'Song title' "
                    "('{}')".format(get_temp_dir() / self.song_file_path),
                ],
            )

            # post assert
            self.assertTrue(vlc_player.playlist_entry_data["song"].started)

            # assert the call
            vlc_player.callbacks["started_song"].assert_called_with(42)

    @patch.object(MediaPlayerVlc, "is_playing_this")
    def test_handle_playing_media_starts_track_id(self, mocked_is_playing_this):
        """Test playing callback when media starts with requested track ID."""
        with self.get_instance() as (vlc_player, (mocked_instance, _, _), _):
            mocked_player = mocked_instance.media_player_new.return_value
            self.set_playlist_entry(vlc_player)
            vlc_player.playlist_entry_data["song"].track_id_audio = 99

            # mock the call
            vlc_player.set_callback("started_song", MagicMock())
            vlc_player.playlist_entry_data["transition"].started = True
            vlc_player.playlist_entry_data["song"].started = False
            mocked_is_playing_this.side_effect = lambda what: what == "song"

            # call the method
            with self.assertLogs("dakara_player.media_player.vlc", "DEBUG") as logger:
                vlc_player.handle_playing("event")

            # assert effect on logs
            self.assertListEqual(
                logger.output,
                [
                    "DEBUG:dakara_player.media_player.vlc:Playing callback called",
                    "DEBUG:dakara_player.media_player.vlc:Requesting to play audio "
                    "track 99",
                    "INFO:dakara_player.media_player.vlc:Now playing 'Song title' "
                    "('{}')".format(get_temp_dir() / self.song_file_path),
                ],
            )

            # assert the call
            vlc_player.callbacks["started_song"].assert_called_with(42)
            mocked_player.audio_set_track.assert_called_with(99)

    @patch.object(MediaPlayerVlc, "is_playing_this")
    def test_handle_playing_idle_starts(self, mocked_is_playing_this):
        """Test playing callback when idle screen starts."""
        with self.get_instance() as (vlc_player, _, _):
            # mock the call
            vlc_player.playlist_entry_data["transition"].started = False
            vlc_player.playlist_entry_data["song"].started = False
            mocked_is_playing_this.side_effect = lambda what: what == "idle"

            # call the method
            with self.assertLogs("dakara_player.media_player.vlc", "DEBUG") as logger:
                vlc_player.handle_playing("event")

            # assert effect on logs
            self.assertListEqual(
                logger.output,
                [
                    "DEBUG:dakara_player.media_player.vlc:Playing callback called",
                    "DEBUG:dakara_player.media_player.vlc:Playing idle screen",
                ],
            )

    @patch.object(MediaPlayerVlc, "is_playing_this")
    def test_handle_playing_invalid(self, mocked_is_playing_this):
        """Test playing callback on invalid state."""
        with self.get_instance() as (vlc_player, _, _):
            # setup mock
            mocked_is_playing_this.return_value = False

            self.assertFalse(vlc_player.stop.is_set())

            # call the method
            with self.assertLogs("dakara_player.media_player.vlc", "DEBUG"):
                vlc_player.handle_playing("event")

            self.assertTrue(vlc_player.stop.is_set())
            exception_class, _, _ = vlc_player.errors.get()
            self.assertIs(InvalidStateError, exception_class)

    @patch.object(MediaPlayerVlc, "get_timing")
    def test_handle_paused(self, mocked_get_timing):
        """Test paused callback."""
        with self.get_instance() as (vlc_player, _, _):
            self.set_playlist_entry(vlc_player)

            # mock the call
            vlc_player.set_callback("paused", MagicMock())
            mocked_get_timing.return_value = 25

            # call the method
            with self.assertLogs("dakara_player.media_player.vlc", "DEBUG") as logger:
                vlc_player.handle_paused("event")

            # assert effect on logs
            self.assertListEqual(
                logger.output,
                [
                    "DEBUG:dakara_player.media_player.vlc:Paused callback called",
                    "DEBUG:dakara_player.media_player.vlc:Paused",
                ],
            )

            # assert the call
            vlc_player.callbacks["paused"].assert_called_with(42, 25)

    @patch.object(PlatformDirs, "user_data_path", new_callable=PropertyMock)
    def test_custom_backgrounds(self, mocked_user_data_path):
        """Test to instanciate with custom backgrounds."""
        mocked_user_data_path.return_value = Path("directory")

        # create object
        tempdir = Path("temp")
        with self.get_instance(
            {
                "backgrounds": {
                    "transition_background_name": "custom_transition.png",
                    "idle_background_name": "custom_idle.png",
                }
            },
            tempdir=tempdir,
        ) as (_, _, (_, mocked_background_loader_class, _)):

            # assert the instanciation of the background loader
            mocked_background_loader_class.assert_called_with(
                destination=tempdir,
                package="dakara_player.resources.backgrounds",
                directory=Path("directory") / "player" / "backgrounds",
                filenames={
                    "transition": "custom_transition.png",
                    "idle": "custom_idle.png",
                },
            )

    def test_default_durations(self):
        """Test to instanciate with default durations."""
        with self.get_instance() as (vlc_player, _, _):
            # assert the instance
            self.assertDictEqual(
                vlc_player.durations,
                {"transition": 10, "idle": 20, "rewind_fast_forward": 10},
            )

    def test_custom_durations(self):
        """Test to instanciate with custom durations."""
        with self.get_instance({"durations": {"transition_duration": 5}}) as (
            vlc_player,
            _,
            _,
        ):
            # assert the instance
            self.assertDictEqual(
                vlc_player.durations,
                {"transition": 5, "idle": 20, "rewind_fast_forward": 10},
            )

    @patch("dakara_player.media_player.base.PLAYER_CLOSING_DURATION", 0)
    @patch.object(MediaPlayerVlc, "stop_player")
    def test_slow_close(self, mocked_stop_player):
        """Test to close VLC when it takes a lot of time."""
        with self.get_instance() as (vlc_player, _, _):
            mocked_stop_player.side_effect = lambda: sleep(1)

            with self.assertLogs("dakara_player.media_player.base", "DEBUG") as logger:
                vlc_player.exit_worker()

            self.assertListEqual(
                logger.output,
                ["WARNING:dakara_player.media_player.base:VLC takes too long to stop"],
            )

    @patch.object(TextGenerator, "get_text")
    def test_generate_text_invalid(self, mocked_get_text):
        """Test to generate invalid text screen."""
        with self.get_instance() as (vlc_player, _, _):
            with self.assertRaisesRegex(
                ValueError, "Unexpected action to generate text for: none"
            ):
                vlc_player.generate_text("none")

        with self.assertRaisesRegex(
            ValueError, "Unexpected action to generate text for: none"
        ):
            vlc_player.generate_text("none")

        mocked_get_text.assert_not_called()

    def test_play_invalid(self):
        """Test to play invalid action."""
        with self.get_instance() as (vlc_player, _, _):
            with self.assertRaisesRegex(ValueError, "Unexpected action to play: none"):
                vlc_player.play("none")

            vlc_player.player.play.assert_not_called()

    def test_set_window_none(self):
        """Test to use default window."""
        with self.get_instance() as (vlc_player, _, _):
            with self.assertLogs("dakara_player.media_player.vlc", "DEBUG") as logger:
                vlc_player.set_window(None)

            self.assertListEqual(
                logger.output,
                ["DEBUG:dakara_player.media_player.vlc:Using VLC default window"],
            )

    @patch("dakara_player.media_player.vlc.platform.system", return_value="Linux")
    def test_set_window_linux(self, mocked_system):
        """Test to use X window."""
        with self.get_instance() as (vlc_player, _, _):
            with self.assertLogs("dakara_player.media_player.vlc", "DEBUG") as logger:
                vlc_player.set_window(99)

            self.assertListEqual(
                logger.output,
                ["DEBUG:dakara_player.media_player.vlc:Associating X window to VLC"],
            )

    @patch("dakara_player.media_player.vlc.platform.system", return_value="Windows")
    def test_set_window_windows(self, mocked_system):
        """Test to use Win API window."""
        with self.get_instance() as (vlc_player, _, _):
            with self.assertLogs("dakara_player.media_player.vlc", "DEBUG") as logger:
                vlc_player.set_window(99)

            self.assertListEqual(
                logger.output,
                [
                    "DEBUG:dakara_player.media_player.vlc:"
                    "Associating Win API window to VLC"
                ],
            )

    @patch("dakara_player.media_player.vlc.platform.system", return_value="other")
    def test_set_window_other(self, mocked_system):
        """Test to set window on unknown platform."""
        with self.get_instance() as (vlc_player, _, _):
            with self.assertRaises(NotImplementedError):
                vlc_player.set_window(99)


@patch("dakara_player.media_player.vlc.METADATA_KEYS_COUNT", 10)
class SetMetadataTestCase(TestCase):
    """Test the set_metadata function."""

    def test_set_first(self):
        """Test to set metadata in first field."""
        media = MagicMock()
        media.get_meta.return_value = None

        set_metadata(media, {"data": "value"})

        media.set_meta.assert_called_with(0, '{"data": "value"}')

    def test_set_second(self):
        """Test to set metadata in second field."""
        media = MagicMock()
        media.get_meta.side_effect = ["value", None]

        set_metadata(media, {"data": "value"})

        media.set_meta.assert_called_with(1, '{"data": "value"}')

    def test_set_fail(self):
        """Test error when unable to set metadata in any field."""
        media = MagicMock()
        media.get_meta.return_value = "value"

        with self.assertRaises(ValueError):
            set_metadata(media, {"data": "value"})


@patch("dakara_player.media_player.vlc.METADATA_KEYS_COUNT", 10)
class GetMetadataTestCase(TestCase):
    """Test the get_metadata function."""

    def test_get_first(self):
        """Test to get metadata from first field."""
        media = MagicMock()
        media.get_meta.return_value = '{"data": "value"}'

        self.assertEqual(get_metadata(media), {"data": "value"})

    def test_get_second_none(self):
        """Test to get metadata from second field with first being none."""
        media = MagicMock()
        media.get_meta.side_effect = [None, '{"data": "value"}']

        self.assertEqual(get_metadata(media), {"data": "value"})

    def test_get_second_text(self):
        """Test to get metadata from second field with first being text."""
        media = MagicMock()
        media.get_meta.side_effect = ["value", '{"data": "value"}']

        self.assertEqual(get_metadata(media), {"data": "value"})

    def test_get_fail(self):
        """Test error when unable to get metadata from any field."""
        media = MagicMock()
        media.get_meta.return_value = "value"

        with self.assertRaises(ValueError):
            get_metadata(media)


@patch.object(MediaPlayerVlc, "is_playing_this")
class OnPlayingThisTestCase(BaseTestCase):
    """Test the decorator for returning early if the player is playing
    something else."""

    def test_in_list(self, mocked_is_playing_this):
        """Test an action in the list of accepted actions."""
        function = MagicMock()
        mocked_is_playing_this.side_effect = lambda what: what == "song"

        function_decorated = on_playing_this(["transition", "song"])(function)

        with self.get_instance() as (player, _, _):
            function_decorated(player)

            function.assert_called_with(player)

    def test_not_in_list(self, mocked_is_playing_this):
        """Test an action not in the list of accepted actions."""
        function = MagicMock()
        mocked_is_playing_this.side_effect = lambda what: what == "idle"

        function_decorated = on_playing_this(["transition", "song"])(function)

        with self.get_instance() as (player, _, _):
            self.assertIsNone(function_decorated(player))

            function.assert_not_called()

    def test_not_in_list_default_return(self, mocked_is_playing_this):
        """Test an action not in the list of accepted actions."""
        function = MagicMock()
        mocked_is_playing_this.side_effect = lambda what: what == "idle"

        function_decorated = on_playing_this(["transition", "song"], 42)(function)

        with self.get_instance() as (player, _, _):
            self.assertEqual(function_decorated(player), 42)



"""Manage background images for media players."""

import logging
from importlib.resources import as_file, files
from pathlib import Path
from shutil import copy

from dakara_base.exceptions import DakaraError

logger = logging.getLogger(__name__)


class BackgroundLoader:
    """Loader for backgrounds.

    It finds background files in a custom directory and fallbacks to a package
    resource. Each background file found is copied to a destination directory.
    If the loader cannot find a file, an exception is raised.

    By instance, given you have the following directory structure:

    - `package.default`
        - `idle.png`
        - `transition.png`
        - `other.png`
    - `/directory/custom`
        - `transition.png`
        - `something.png`

    and you use the following configuration:
    >>> from pathlib import Path
    >>> loader = BackgroundLoader(
    ...        destination=Path("/destination"),
    ...        directory=Path("/directory/custom"),
    ...        package="package.default",
    ...        filenames={
    ...            "idle": "idle.png",
    ...            "transition": "transition.png",
    ...            "other": "something.png"
    ...        }
    ...    )

    you will have the following result:
    >>> loader.load()
    >>> loader.backgrounds
        {
            "idle": Path("/destination/idle.png"),
            "transition": Path("/destination/transition.png"),
            "other": Path("/destination/something.png")
        }

    Args:
        destination (pathlib.Path): Where to copy found background files.
        package (str): Package checked for backgrounds by default.
        directory (pathlib.Path): Custom directory checked for backgrounds.
        filenames (dict): Dictionary of background filenames. The key is the
            background name, the value the background file name.

    Attributes:
        backgrounds (dict): Dictionary of background file paths. The key is
            the background name, the value the background file path.
        destination (pathlib.Path): Where to copy found background files.
        package (str): Package checked for backgrounds by default.
        directory (pathlib.Path): Custom directory checked for backgrounds.
        filenames (dict): Dictionary of background filenames. The key is the
            background name, the value the background file name.
    """

    def __init__(
        self,
        destination,
        package,
        directory=None,
        filenames=None,
    ):
        self.destination = destination
        self.package = package
        self.directory = directory
        self.filenames = filenames or {}
        self.backgrounds = {}

    def load(self):
        """Load the backgrounds."""
        logger.debug("Loading backgrounds")
        for name, file_name in self.filenames.items():
            self.backgrounds[name] = self.get_background_path(name, file_name)

    def get_background_path(self, background_name, file_name):
        """Get the accurate path of one background

        Args:
            background_name (str): Name of the background.
            file_name (str): Name of the background file.

        Returns:
            pathlib.Path: Absolute path to the background file.
        """
        # trying to load from custom directory
        if self.directory is not None:
            try:
                return self.copy_custom_background(background_name, file_name)

            except FileNotFoundError:
                pass

        # trying to load from package by default
        try:
            return self.copy_default_background(background_name, file_name)

        except FileNotFoundError as error:
            raise BackgroundNotFoundError(
                f"No {background_name} background file found for '{file_name}'"
            ) from error

    def copy_custom_background(self, background_name, file_name):
        """Copy a custom background.

        Args:
            background_name (str): Name of the background.
            file_name (str): Name of the background file.

        Returns:
            pathlib.Path: Absolute path to the background file.

        Raises:
            FileNotFoundError: If the custom file does not exist (by
                `shuti.copy`).
        """
        file_path = self.directory / file_name
        output_path = Path(copy(file_path, self.destination))
        logger.debug(
            "Loading custom %s background file '%s'", background_name, file_name
        )
        return output_path

    def copy_default_background(self, background_name, file_name):
        """Copy a default background.

        Args:
            background_name (str): Name of the background.
            file_name (str): Name of the background file.

        Returns:
            pathlib.Path: Absolute path to the background file.

        Raises:
            FileNotFoundError: If the custom file does not exist (by
                `shuti.copy`).
        """
        with as_file(files(self.package).joinpath(file_name)) as file_path:
            output_path = Path(copy(file_path, self.destination))
            logger.debug(
                "Loading default %s background file '%s'",
                background_name,
                file_name,
            )
            return output_path


class BackgroundNotFoundError(DakaraError, FileNotFoundError):
    """Error raised when a background cannot be found"""



from contextlib import ExitStack, contextmanager
from pathlib import Path
from queue import Queue
from tempfile import TemporaryDirectory
from threading import Event
from time import sleep
from unittest import skipUnless
from unittest.mock import MagicMock

from dakara_base.config import Config
from func_timeout import func_set_timeout

from dakara_player.media_player.base import (
    IDLE_BG_NAME,
    IDLE_TEXT_NAME,
    TRANSITION_BG_NAME,
    TRANSITION_TEXT_NAME,
)
from dakara_player.media_player.mpv import MediaPlayerMpv
from tests.integration.base import TestCasePollerKara

REWIND_FAST_FORWARD_DURATION = 0.5
REWIND_FAST_FORWARD_DELTA = 1
DEFAULT_DELTA = 0.2


@skipUnless(MediaPlayerMpv.is_available(), "mpv not installed")
class MediaPlayerMpvIntegrationTestCase(TestCasePollerKara):
    """Test the mpv player class in real conditions."""

    TIMEOUT = 120
    DELAY = 0.2

    def setUp(self):
        super().setUp()

        # create fullscreen flag
        self.fullscreen = True

        # create transition duration
        self.transition_duration = 1

    @contextmanager
    def get_instance(self, config=None, check_error=True):
        """Get an instance of MediaPlayerMpv for the available version.

        This method is a context manager that automatically stops the player on
        exit.

        Args:
            config (dict): Extra configuration passed to the constructor.
            check_error (bool): If true, check if the player stop event is not
                set and the error queue is empty at the end.

        Yields:
            tuple: Containing the following elements:
                MediaPlayerMpv: Instance;
                pathlib.Path: Path of the temporary directory;
                unittest.case._LoggingWatcher: Captured output.
        """
        config_full = {
            "kara_folder": str(self.kara_folder_path),
            "fullscreen": self.fullscreen,
            "mpv": {"vo": "null", "ao": "null"},
        }

        if config:
            config_full.update(config)

        with TemporaryDirectory() as temp_str:
            temp = Path(temp_str).resolve()
            try:
                with ExitStack() as stack:
                    mpv_player = stack.enter_context(
                        MediaPlayerMpv.from_version(
                            Event(),
                            Queue(),
                            Config("DAKARA_PLAYER", config_full),
                            temp,
                            warn_long_exit=False,
                        )
                    )
                    output = stack.enter_context(
                        self.assertLogs("dakara_player.media_player.mpv", "DEBUG")
                    )
                    mpv_player.load()

                    yield mpv_player, temp, output

                    if check_error:
                        # display errors in queue if any
                        if not mpv_player.errors.empty():
                            _, error, traceback = mpv_player.errors.get(5)
                            error.with_traceback(traceback)
                            raise error

                        # assert no errors to fail test if any
                        self.assertFalse(mpv_player.stop.is_set())

            except OSError:
                # silence closing errors of mpv
                pass

            # sleep to allow slow systems to correctly clean up
            sleep(self.DELAY)

    @func_set_timeout(TIMEOUT)
    def test_start(self):
        """Test the initial state of the player without instructions."""
        with self.get_instance() as (mpv_player, _, _):
            self.assertIsNone(mpv_player.playlist_entry)
            self.assertIsNone(mpv_player.playlist_entry_data["transition"].path)
            self.assertIsNone(mpv_player.playlist_entry_data["song"].path)
            self.assertFalse(mpv_player.is_playing_this("idle"))
            self.assertFalse(mpv_player.is_playing_this("transition"))
            self.assertFalse(mpv_player.is_playing_this("song"))

    @func_set_timeout(TIMEOUT)
    def test_play_idle(self):
        """Test to display the idle screen."""
        with self.get_instance() as (mpv_player, temp, _):
            # pre assertions
            self.assertIsNone(mpv_player.player.path)

            # call the method
            mpv_player.play("idle")

            # wait for the idle screen to start
            self.wait_is_playing(mpv_player, "idle")

            # post assertions
            self.assertIsNotNone(mpv_player.player.path)
            self.assertEqual(mpv_player.player.path, str(temp / IDLE_BG_NAME))
            self.assertListEqual(
                mpv_player.player.sub_files, [str(temp / IDLE_TEXT_NAME)]
            )

    @func_set_timeout(TIMEOUT)
    def test_play_playlist_entry(self):
        """Test to play a playlist entry.

        First, the transition screen is played, then the song itself.
        """
        with self.get_instance() as (mpv_player, temp, _):
            # mock the callbacks
            mpv_player.set_callback("started_transition", MagicMock())
            mpv_player.set_callback("started_song", MagicMock())

            # pre assertions
            self.assertIsNone(mpv_player.playlist_entry)
            self.assertIsNone(mpv_player.player.path)

            # set playlist entry
            mpv_player.set_playlist_entry(self.playlist_entry1, autoplay=False)

            # check memory
            self.assertEqual(
                mpv_player.playlist_entry_data["transition"].path,
                temp / TRANSITION_BG_NAME,
            )
            self.assertEqual(
                mpv_player.playlist_entry_data["song"].path, self.song1_path
            )
            self.assertEqual(
                mpv_player.playlist_entry_data["song"].path_subtitle,
                self.subtitle1_path,
            )

            # start playing
            mpv_player.play("transition")

            # wait for the transition screen to start
            self.wait_is_playing(mpv_player, "transition")

            # post assertions for transition screen
            self.assertIsNotNone(mpv_player.playlist_entry)

            # check media
            self.assertIsNotNone(mpv_player.player.path)
            self.assertEqual(mpv_player.player.path, str(temp / TRANSITION_BG_NAME))

            # check there is no audio track
            self.assertFalse(mpv_player.player.audio)

            # check which subtitle file is read
            self.assertListEqual(
                mpv_player.player.sub_files, [str(temp / TRANSITION_TEXT_NAME)]
            )

            # assert the started transition callback has been called
            mpv_player.callbacks["started_transition"].assert_called_with(
                self.playlist_entry1["id"]
            )

            # wait for the media to start
            self.wait_is_playing(mpv_player, "song")

            # check media
            self.assertIsNotNone(mpv_player.player.path)
            self.assertEqual(mpv_player.player.path, str(self.song1_path))

            # check audio track
            self.assertEqual(mpv_player.player.audio, 1)

            # check there is no audio file
            self.assertEqual(len(mpv_player.player.audio_files), 0)

            # assert the started song callback has been called
            mpv_player.callbacks["started_song"].assert_called_with(
                self.playlist_entry1["id"]
            )

            # assert the player is playing a song
            self.assertFalse(mpv_player.is_playing_this("transition"))
            self.assertTrue(mpv_player.is_playing_this("song"))
            self.assertFalse(mpv_player.is_playing_this("idle"))

            # wait for the media to end
            self.wait(mpv_player, lambda: not mpv_player.is_playing())

            # assert the player is not playing anything
            self.assertFalse(mpv_player.is_playing_this("transition"))
            self.assertFalse(mpv_player.is_playing_this("song"))
            self.assertFalse(mpv_player.is_playing_this("idle"))

    @func_set_timeout(TIMEOUT)
    def test_play_playlist_entry_instrumental_track(self):
        """Test to play a playlist entry using instrumental track."""
        # request to use instrumental track
        self.playlist_entry1["use_instrumental"] = True

        with self.get_instance() as (mpv_player, _, _):
            # mock the callbacks
            mpv_player.set_callback("started_transition", MagicMock())
            mpv_player.set_callback("started_song", MagicMock())

            # pre assertions
            self.assertIsNone(mpv_player.playlist_entry)
            self.assertIsNone(mpv_player.player.path)

            # call the method
            mpv_player.set_playlist_entry(self.playlist_entry1)

            # wait for the song to start
            self.wait_is_playing(mpv_player, "song")

            # check the current media has 2 audio tracks
            self.assertListEqual(mpv_player.get_track_id_audio_list(), [1, 2])

            # check media exists
            self.assertIsNotNone(mpv_player.player.path)
            self.assertEqual(mpv_player.player.path, str(self.song1_path))

            # check audio track
            self.assertEqual(mpv_player.player.audio, 2)

            # check there is no audio file
            self.assertEqual(len(mpv_player.player.audio_files), 0)

            # assert the started song callback has been called
            mpv_player.callbacks["started_song"].assert_called_with(
                self.playlist_entry1["id"]
            )

    @func_set_timeout(TIMEOUT)
    def test_play_playlist_entry_instrumental_file(self):
        """Test to play a playlist entry using instrumental file."""
        # request to use instrumental file
        self.playlist_entry1["song"]["file_path"] = self.song2_path
        self.playlist_entry1["use_instrumental"] = True

        with self.get_instance() as (mpv_player, _, _):
            # mock the callbacks
            mpv_player.set_callback("started_transition", MagicMock())
            mpv_player.set_callback("started_song", MagicMock())

            # pre assertions
            self.assertIsNone(mpv_player.playlist_entry)
            self.assertIsNone(mpv_player.player.path)

            # call the method
            mpv_player.set_playlist_entry(self.playlist_entry1)

            # wait for the song to start
            self.wait_is_playing(mpv_player, "song")

            # check media exists
            self.assertIsNotNone(mpv_player.player.path)
            self.assertEqual(mpv_player.player.path, str(self.song2_path))

            # check audio track
            self.assertEqual(mpv_player.player.audio, 3)

            # check audio file
            self.assertEqual(len(mpv_player.player.audio_files), 1)
            self.assertEqual(mpv_player.player.audio_files[0], str(self.audio2_path))

            # assert the started song callback has been called
            mpv_player.callbacks["started_song"].assert_called_with(
                self.playlist_entry1["id"]
            )

    @func_set_timeout(TIMEOUT)
    def test_pause(self):
        """Test to pause and unpause the player."""
        with self.get_instance() as (mpv_player, _, _):
            # mock the callbacks
            mpv_player.set_callback("paused", MagicMock())
            mpv_player.set_callback("resumed", MagicMock())

            # start the playlist entry
            mpv_player.set_playlist_entry(self.playlist_entry1)

            # wait for the song to start
            self.wait_is_playing(mpv_player, "song")

            # call the method to pause the player
            mpv_player.pause()
            timing = mpv_player.get_timing()

            # wait for the player to be paused
            self.wait_is_paused(mpv_player)

            # assert in pause
            self.assertFalse(mpv_player.is_playing())

            # assert the callback
            mpv_player.callbacks["paused"].assert_called_with(
                self.playlist_entry1["id"], timing
            )
            mpv_player.callbacks["resumed"].assert_not_called()

            # reset the mocks
            mpv_player.callbacks["paused"].reset_mock()
            mpv_player.callbacks["resumed"].reset_mock()

            # call the method to resume the player
            mpv_player.resume()

            # wait for the player to play again
            self.wait_is_playing(mpv_player)

            # assert the callback
            mpv_player.callbacks["paused"].assert_not_called()
            mpv_player.callbacks["resumed"].assert_called_with(
                self.playlist_entry1["id"],
                timing,  # on a slow computer, the timing may be inaccurate
            )

    @func_set_timeout(TIMEOUT)
    def test_double_pause(self):
        """Test that double pause and double resume have no effects."""
        with self.get_instance() as (mpv_player, _, _):
            # mock the callbacks
            mpv_player.set_callback("paused", MagicMock())
            mpv_player.set_callback("resumed", MagicMock())

            # start the playlist entry
            mpv_player.set_playlist_entry(self.playlist_entry1)

            # wait for the song to start
            self.wait_is_playing(mpv_player, "song")

            # call the method to pause the player
            mpv_player.pause()

            # wait for the player to be paused
            self.wait_is_paused(mpv_player)

            # assert the callback
            mpv_player.callbacks["paused"].assert_called()
            mpv_player.callbacks["resumed"].assert_not_called()

            # reset the mocks
            mpv_player.callbacks["paused"].reset_mock()
            mpv_player.callbacks["resumed"].reset_mock()

            # re-call the method to pause the player
            mpv_player.pause()

            # wait again for the player to be paused
            self.wait_is_paused(mpv_player)

            # assert the callback
            mpv_player.callbacks["paused"].assert_not_called()
            mpv_player.callbacks["resumed"].assert_not_called()

            # reset the mocks
            mpv_player.callbacks["paused"].reset_mock()
            mpv_player.callbacks["resumed"].reset_mock()

            # call the method to resume the player
            mpv_player.resume()

            # wait for the player to play again
            self.wait_is_playing(mpv_player)

            # assert the callback
            mpv_player.callbacks["paused"].assert_not_called()
            mpv_player.callbacks["resumed"].assert_called()

            # reset the mocks
            mpv_player.callbacks["paused"].reset_mock()
            mpv_player.callbacks["resumed"].reset_mock()

            # re-call the method to resume the player
            mpv_player.resume()
            self.wait_is_playing(mpv_player)

            # assert the callback
            mpv_player.callbacks["paused"].assert_not_called()
            mpv_player.callbacks["resumed"].assert_not_called()

    @func_set_timeout(TIMEOUT)
    def test_restart_song(self):
        """Test to restart a playlist entry."""
        with self.get_instance() as (mpv_player, _, _):
            # mock the callbacks
            mpv_player.set_callback("started_transition", MagicMock())
            mpv_player.set_callback("started_song", MagicMock())
            mpv_player.set_callback("finished", MagicMock())
            mpv_player.set_callback("updated_timing", MagicMock())

            # pre assertions
            self.assertIsNone(mpv_player.playlist_entry)
            self.assertIsNone(mpv_player.player.path)

            # request playlist entry to play
            mpv_player.set_playlist_entry(self.playlist_entry1)

            # wait for the media to start
            self.wait_is_playing(mpv_player, "song")

            # post assertions for song
            self.assertIsNotNone(mpv_player.playlist_entry)

            # wait a bit for the player to play
            self.wait(
                mpv_player,
                lambda: mpv_player.player.time_pos >= REWIND_FAST_FORWARD_DURATION,
            )

            # request playlist entry to restart
            mpv_player.restart()

            # check timing is 0
            self.assertAlmostEqual(mpv_player.player.time_pos, 0, delta=DEFAULT_DELTA)

            # check the song is not stopped
            self.assertIsNotNone(mpv_player.playlist_entry)
            mpv_player.callbacks["finished"].assert_not_called()

            # assert callback
            mpv_player.callbacks["updated_timing"].assert_called_with(
                self.playlist_entry1["id"], 0
            )

    @func_set_timeout(TIMEOUT)
    def test_skip_song(self):
        """Test to skip a playlist entry."""
        with self.get_instance() as (mpv_player, _, _):
            # mock the callbacks
            mpv_player.set_callback("started_transition", MagicMock())
            mpv_player.set_callback("started_song", MagicMock())
            mpv_player.set_callback("finished", MagicMock())

            # pre assertions
            self.assertIsNone(mpv_player.playlist_entry)
            self.assertIsNone(mpv_player.player.path)

            # request initial playlist entry to play
            mpv_player.set_playlist_entry(self.playlist_entry1)

            # wait for the media to start
            self.wait_is_playing(mpv_player, "song")

            # post assertions for song
            self.assertIsNotNone(mpv_player.playlist_entry)

            # check media
            self.assertIsNotNone(mpv_player.player.path)
            self.assertEqual(mpv_player.player.path, str(self.song1_path))

            # request first playlist entry to stop
            mpv_player.skip()

            # check skip flag
            self.assertTrue(mpv_player.player_data["skip"])

            # check the song is stopped accordingly
            self.assertIsNone(mpv_player.playlist_entry)
            mpv_player.callbacks["finished"].assert_called_with(
                self.playlist_entry1["id"]
            )

            # request second playlist entry to play
            mpv_player.set_playlist_entry(self.playlist_entry2)

            # wait for the media to start
            self.wait_is_playing(mpv_player, "song")

            # post assertions for song
            self.assertIsNotNone(mpv_player.playlist_entry)

            # check media
            self.assertIsNotNone(mpv_player.player.path)
            self.assertEqual(mpv_player.player.path, str(self.song2_path))

            # check skip flag
            self.assertFalse(mpv_player.player_data["skip"])

    @func_set_timeout(TIMEOUT)
    def test_skip_last_song(self):
        """Test to skip the last playlist entry."""
        with self.get_instance() as (mpv_player, _, _):
            # mock the callbacks
            mpv_player.set_callback("started_transition", MagicMock())
            mpv_player.set_callback("started_song", MagicMock())
            mpv_player.set_callback("finished", MagicMock())

            # pre assertions
            self.assertIsNone(mpv_player.playlist_entry)
            self.assertIsNone(mpv_player.player.path)

            # request initial playlist entry to play
            mpv_player.set_playlist_entry(self.playlist_entry1)

            # wait for the media to start
            self.wait_is_playing(mpv_player, "song")

            # post assertions for song
            self.assertIsNotNone(mpv_player.playlist_entry)

            # check media
            self.assertIsNotNone(mpv_player.player.path)
            self.assertEqual(mpv_player.player.path, str(self.song1_path))

            # request first playlist entry to stop
            mpv_player.skip()

            # check skip flag
            self.assertTrue(mpv_player.player_data["skip"])

            # check the song is stopped accordingly
            self.assertIsNone(mpv_player.playlist_entry)
            mpv_player.callbacks["finished"].assert_called_with(
                self.playlist_entry1["id"]
            )

            # request idle screen
            mpv_player.play("idle")

            # wait for the media to start
            self.wait_is_playing(mpv_player, "idle")

            # check skip flag
            self.assertFalse(mpv_player.player_data["skip"])

    @func_set_timeout(TIMEOUT)
    def test_skip_song_pause(self):
        """Test to skip playlist entry on pause."""
        with self.get_instance() as (mpv_player, _, _):
            # mock the callbacks
            mpv_player.set_callback("started_transition", MagicMock())
            mpv_player.set_callback("started_song", MagicMock())
            mpv_player.set_callback("finished", MagicMock())
            mpv_player.set_callback("pause", MagicMock())
            mpv_player.set_callback("resumed", MagicMock())

            # pre assertions
            self.assertIsNone(mpv_player.playlist_entry)
            self.assertIsNone(mpv_player.player.path)

            # request initial playlist entry to play
            mpv_player.set_playlist_entry(self.playlist_entry1)

            # wait for the media to start
            self.wait_is_playing(mpv_player, "song")

            # post assertions for song
            self.assertIsNotNone(mpv_player.playlist_entry)

            # check media
            self.assertIsNotNone(mpv_player.player.path)
            self.assertEqual(mpv_player.player.path, str(self.song1_path))

            # pause song
            mpv_player.pause()

            # wait for the media to be paused
            self.wait_is_paused(mpv_player)

            # request first playlist entry to stop
            mpv_player.skip()

            # check skip flag
            self.assertTrue(mpv_player.player_data["skip"])

            # check the song is stopped accordingly
            self.assertIsNone(mpv_player.playlist_entry)
            mpv_player.callbacks["finished"].assert_called_with(
                self.playlist_entry1["id"]
            )

            # request second playlist entry to play
            mpv_player.set_playlist_entry(self.playlist_entry2)

            # wait for the media to start
            self.wait_is_playing(mpv_player, "song")

            # post assertions for song
            self.assertIsNotNone(mpv_player.playlist_entry)

            # check media
            self.assertIsNotNone(mpv_player.player.path)
            self.assertEqual(mpv_player.player.path, str(self.song2_path))

            # check skip flag
            self.assertFalse(mpv_player.player_data["skip"])

            # check callbacks
            mpv_player.callbacks["resumed"].assert_not_called()

    @func_set_timeout(TIMEOUT)
    def test_skip_last_song_pause(self):
        """Test to skip the last playlist entry on pause."""
        with self.get_instance() as (mpv_player, _, _):
            # mock the callbacks
            mpv_player.set_callback("started_transition", MagicMock())
            mpv_player.set_callback("started_song", MagicMock())
            mpv_player.set_callback("finished", MagicMock())
            mpv_player.set_callback("pause", MagicMock())
            mpv_player.set_callback("resumed", MagicMock())

            # pre assertions
            self.assertIsNone(mpv_player.playlist_entry)
            self.assertIsNone(mpv_player.player.path)

            # request initial playlist entry to play
            mpv_player.set_playlist_entry(self.playlist_entry1)

            # wait for the media to start
            self.wait_is_playing(mpv_player, "song")

            # post assertions for song
            self.assertIsNotNone(mpv_player.playlist_entry)

            # check media
            self.assertIsNotNone(mpv_player.player.path)
            self.assertEqual(mpv_player.player.path, str(self.song1_path))

            # pause song
            mpv_player.pause()

            # wait for the media to be paused
            self.wait_is_paused(mpv_player)

            # request first playlist entry to stop
            mpv_player.skip()

            # check the song is stopped accordingly
            self.assertIsNone(mpv_player.playlist_entry)
            mpv_player.callbacks["finished"].assert_called_with(
                self.playlist_entry1["id"]
            )

            # request idle screen
            mpv_player.play("idle")

            # wait for the media to start
            self.wait_is_playing(mpv_player, "idle")

            # check callbacks
            mpv_player.callbacks["resumed"].assert_not_called()

    @func_set_timeout(TIMEOUT)
    def test_skip_transition(self):
        """Test to skip a playlist entry transition screen."""
        with self.get_instance() as (mpv_player, _, _):
            # mock the callbacks
            mpv_player.set_callback("started_transition", MagicMock())
            mpv_player.set_callback("started_song", MagicMock())
            mpv_player.set_callback("finished", MagicMock())

            # pre assertions
            self.assertIsNone(mpv_player.playlist_entry)
            self.assertIsNone(mpv_player.player.path)

            # request initial playlist entry to play
            mpv_player.set_playlist_entry(self.playlist_entry1)

            # wait for the transition to start
            self.wait_is_playing(mpv_player, "transition")

            # request first playlist entry to stop
            mpv_player.skip()

            # check the song is stopped accordingly
            mpv_player.callbacks["finished"].assert_called_with(
                self.playlist_entry1["id"]
            )

            # request second playlist entry to play
            mpv_player.set_playlist_entry(self.playlist_entry2)

            # wait for the media to start
            self.wait_is_playing(mpv_player, "song")

            # check media
            self.assertIsNotNone(mpv_player.player.path)
            self.assertEqual(mpv_player.player.path, str(self.song2_path))

    @func_set_timeout(TIMEOUT)
    def test_skip_idle(self):
        """Test to play a playlist entry after idle screen."""
        with self.get_instance() as (mpv_player, _, _):
            # mock the callbacks
            mpv_player.set_callback("started_transition", MagicMock())
            mpv_player.set_callback("started_song", MagicMock())
            mpv_player.set_callback("finished", MagicMock())

            # pre assertions
            self.assertIsNone(mpv_player.playlist_entry)
            self.assertIsNone(mpv_player.player.path)

            # request idle screen
            mpv_player.play("idle")

            # wait for the media to start
            self.wait_is_playing(mpv_player, "idle")

            # post assertions for song
            self.assertIsNone(mpv_player.playlist_entry)

            # request playlist entry to play
            mpv_player.set_playlist_entry(self.playlist_entry1)

            # wait for the media to start
            self.wait_is_playing(mpv_player, "song")

            # post assertions for song
            self.assertIsNotNone(mpv_player.playlist_entry)

            # check media
            self.assertIsNotNone(mpv_player.player.path)
            self.assertEqual(mpv_player.player.path, str(self.song1_path))

    @func_set_timeout(TIMEOUT)
    def test_rewind_song(self):
        """Test to rewind a playlist entry."""
        with self.get_instance(
            {
                "durations": {
                    "rewind_fast_forward_duration": REWIND_FAST_FORWARD_DURATION
                }
            }
        ) as (mpv_player, _, _):
            # mock the callbacks
            mpv_player.set_callback("started_transition", MagicMock())
            mpv_player.set_callback("started_song", MagicMock())
            mpv_player.set_callback("updated_timing", MagicMock())

            # pre assertions
            self.assertIsNone(mpv_player.playlist_entry)
            self.assertIsNone(mpv_player.player.path)

            # request playlist entry to play
            mpv_player.set_playlist_entry(self.playlist_entry1)

            # wait for the media to start
            self.wait_is_playing(mpv_player, "song")

            # post assertions for song
            self.assertIsNotNone(mpv_player.playlist_entry)

            # wait a bit for the player to play
            self.wait(
                mpv_player,
                lambda: mpv_player.player.time_pos >= REWIND_FAST_FORWARD_DURATION * 2,
            )
            timing1 = mpv_player.player.time_pos

            # request playlist entry to rewind
            mpv_player.rewind()

            # check timing is earlier than previously
            timing2 = mpv_player.player.time_pos
            self.assertLess(timing2, timing1)
            self.assertAlmostEqual(
                timing1 - timing2,
                REWIND_FAST_FORWARD_DURATION,
                delta=REWIND_FAST_FORWARD_DELTA,
            )

            # assert callback
            mpv_player.callbacks["updated_timing"].assert_called_with(
                self.playlist_entry1["id"], int(timing2)
            )

    @func_set_timeout(TIMEOUT)
    def test_rewind_song_before_start(self):
        """Test to rewind a playlist entry before its start."""
        with self.get_instance() as (mpv_player, _, _):
            # mock the callbacks
            mpv_player.set_callback("started_transition", MagicMock())
            mpv_player.set_callback("started_song", MagicMock())

            # pre assertions
            self.assertIsNone(mpv_player.playlist_entry)
            self.assertIsNone(mpv_player.player.path)

            # request playlist entry to play
            mpv_player.set_playlist_entry(self.playlist_entry1)

            # wait for the media to start
            self.wait_is_playing(mpv_player, "song")

            # post assertions for song
            self.assertIsNotNone(mpv_player.playlist_entry)

            # request playlist entry to rewind
            mpv_player.rewind()

            # check timing is 0
            self.assertAlmostEqual(mpv_player.player.time_pos, 0, delta=DEFAULT_DELTA)

    @func_set_timeout(TIMEOUT)
    def test_fast_forward_song(self):
        """Test to advance a playlist entry."""
        with self.get_instance(
            {
                "durations": {
                    "rewind_fast_forward_duration": REWIND_FAST_FORWARD_DURATION
                }
            }
        ) as (mpv_player, _, _):
            # mock the callbacks
            mpv_player.set_callback("started_transition", MagicMock())
            mpv_player.set_callback("started_song", MagicMock())
            mpv_player.set_callback("updated_timing", MagicMock())

            # pre assertions
            self.assertIsNone(mpv_player.playlist_entry)
            self.assertIsNone(mpv_player.player.path)

            # request playlist entry to play
            mpv_player.set_playlist_entry(self.playlist_entry1)

            # wait for the media to start
            self.wait_is_playing(mpv_player, "song")

            # post assertions for song
            self.assertIsNotNone(mpv_player.playlist_entry)

            # wait a bit for the player to play
            self.wait(
                mpv_player,
                lambda: mpv_player.player.time_pos >= REWIND_FAST_FORWARD_DURATION * 2,
            )
            timing1 = mpv_player.player.time_pos

            # request playlist entry to advance
            mpv_player.fast_forward()

            # check timing is later than previously
            timing2 = mpv_player.player.time_pos
            self.assertGreater(timing2, timing1)
            self.assertAlmostEqual(
                timing2 - timing1,
                REWIND_FAST_FORWARD_DURATION,
                delta=REWIND_FAST_FORWARD_DELTA,
            )

            # assert callback
            mpv_player.callbacks["updated_timing"].assert_called_with(
                self.playlist_entry1["id"], int(timing2)
            )

    @func_set_timeout(TIMEOUT)
    def test_fast_forward_song_after_end(self):
        """Test to advance a playlist entry after its end."""
        with self.get_instance() as (mpv_player, _, _):
            # mock the callbacks
            mpv_player.set_callback("started_transition", MagicMock())
            mpv_player.set_callback("started_song", MagicMock())
            mpv_player.set_callback("finished", MagicMock())

            # pre assertions
            self.assertIsNone(mpv_player.playlist_entry)
            self.assertIsNone(mpv_player.player.path)

            # request playlist entry to play
            mpv_player.set_playlist_entry(self.playlist_entry1)

            # wait for the media to start
            self.wait_is_playing(mpv_player, "song")

            # post assertions for song
            self.assertIsNotNone(mpv_player.playlist_entry)

            # request playlist entry to advance
            mpv_player.fast_forward()

            # check the song has finished
            mpv_player.callbacks["finished"].assert_called_with(
                self.playlist_entry1["id"]
            )

    @func_set_timeout(TIMEOUT)
    def test_play_idle_after_playlist_paused(self):
        """Test to request idle screen when playing a playlist entry on pause."""
        with self.get_instance() as (mpv_player, _, _):
            # mock the callbacks
            mpv_player.set_callback("started_transition", MagicMock())
            mpv_player.set_callback("started_song", MagicMock())
            mpv_player.set_callback("finished", MagicMock())
            mpv_player.set_callback("pause", MagicMock())
            mpv_player.set_callback("resumed", MagicMock())

            # pre assertions
            self.assertIsNone(mpv_player.playlist_entry)
            self.assertIsNone(mpv_player.player.path)

            # request initial playlist entry to play
            mpv_player.set_playlist_entry(self.playlist_entry1)

            # wait for the media to start
            self.wait_is_playing(mpv_player, "song")

            # post assertions for song
            self.assertIsNotNone(mpv_player.playlist_entry)

            # check media
            self.assertIsNotNone(mpv_player.player.path)
            self.assertEqual(mpv_player.player.path, str(self.song1_path))

            # pause song
            mpv_player.pause()

            # wait for the media to be paused
            self.wait_is_paused(mpv_player)

            mpv_player.skip(no_callback=True)

            # call the method
            mpv_player.play("idle")

            # wait for the idle screen to start
            self.wait_is_playing(mpv_player, "idle")

            # post assertions
            self.assertIsNotNone(mpv_player.player.path)

            # assert the call
            mpv_player.callbacks["resumed"].assert_not_called()



"""Manage the user resource directory and files."""

import logging
from importlib.resources import files
from shutil import copy

from dakara_base.directory import directories
from dakara_base.utils import strtobool

logger = logging.getLogger(__name__)


def copy_resource(package, destination, force):
    """Copy the content of one resource directory.

    Args:
        package (str): Package of resources to copy.
        destination (pathlib.Path): Directory where to copy the resource.
        force (bool): If the destination exists and this flag is set to `True`,
            overwrite the destination.
    """
    if not force and destination.exists():
        result = strtobool(
            input(
                f"Directory {destination} already exists, "
                "overwrite it with its content? [y/N] "
            )
        )

        if not result:
            return

    destination.mkdir(parents=True, exist_ok=True)

    for resource in files(package).iterdir():
        # ignore Python files
        if resource.name.startswith("__"):
            continue

        copy(resource, destination)


def create_resource_files(force=False):
    """Copy the resource files to user directory.

    Args:
        force (bool): If the user directory already contains the resource
            directories and this flag is set, overwrite the directories.
    """
    user_directory = directories.user_data_path
    user_directory.mkdir(parents=True, exist_ok=True)

    for directory in ["backgrounds", "templates"]:
        copy_resource(
            f"dakara_player.resources.{directory}", user_directory / directory, force
        )

    logger.info(f"Resource files created in '{user_directory}'")



from queue import Queue
from threading import Event
from unittest import TestCase
from unittest.mock import ANY, MagicMock, patch

from dakara_player.web_client import HTTPClientDakara, WebSocketClientDakara


class HTTPClientDakaraTestCase(TestCase):
    """Test the HTTP client."""

    def setUp(self):
        # create a token
        self.token = "token value"

        # create a server address
        self.address = "www.example.com"

        # create a server URL
        self.url = "http://www.example.com/api"

        # create a login and password
        self.login = "test"
        self.password = "test"

        # create a HTTPClientDakara instance
        self.http_client = HTTPClientDakara(
            {"address": self.address, "login": self.login, "password": self.password},
            endpoint_prefix="api",
        )
        self.set_token()

    def set_token(self):
        """Set the token for the test client."""
        self.http_client.token = "Token"

    def test_init_url(self):
        """Test the object has the correct URL."""
        self.assertEqual(self.http_client.server_url, self.url)

    @patch.object(HTTPClientDakara, "post")
    def test_post_player_error_successful(self, mocked_post):
        """Test to report an error sucessfuly."""
        # call the method
        with self.assertLogs("dakara_player.web_client", "DEBUG") as logger:
            self.http_client.post_player_error(42, "message")

        # assert the effect on logs
        self.assertListEqual(
            logger.output,
            [
                "DEBUG:dakara_player.web_client:"
                "Telling the server that playlist entry 42 cannot be played"
            ],
        )

        # assert the call
        mocked_post.assert_called_with(
            endpoint="playlist/player/errors/",
            data={"playlist_entry_id": 42, "error_message": "message"},
            message_on_error="Unable to send player error to server",
        )

    @patch.object(HTTPClientDakara, "post")
    def test_post_player_error_failed(self, mocked_post):
        """Test to report an invalid error."""
        # call the method
        with self.assertRaises(AssertionError):
            self.http_client.post_player_error(None, "message")

        # assert the call
        mocked_post.assert_not_called()

    @patch.object(HTTPClientDakara, "put")
    def test_put_status_finished_successful(self, mocked_put):
        """Test to report that a playlist entry finished."""
        # call the method
        with self.assertLogs("dakara_player.web_client", "DEBUG") as logger:
            self.http_client.put_status_finished(42)

        # assert the effect on logs
        self.assertListEqual(
            logger.output,
            [
                "DEBUG:dakara_player.web_client:"
                "Telling the server that playlist entry 42 is finished"
            ],
        )

        # assert the call
        mocked_put.assert_called_with(
            endpoint="playlist/player/status/",
            data={"event": "finished", "playlist_entry_id": 42},
            message_on_error="Unable to report that a playlist entry has finished",
        )

    @patch.object(HTTPClientDakara, "put")
    def test_put_status_finished_failed(self, mocked_put):
        """Test to report that an invalid playlist entry finished."""
        # call the method
        with self.assertRaises(AssertionError):
            self.http_client.put_status_finished(None)

        # assert the call
        mocked_put.assert_not_called()

    @patch.object(HTTPClientDakara, "put")
    def test_put_status_started_transition_successful(self, mocked_put):
        """Test to report that a playlist entry transition started."""
        # call the method
        with self.assertLogs("dakara_player.web_client", "DEBUG") as logger:
            self.http_client.put_status_started_transition(42)

        # assert the effect on logs
        self.assertListEqual(
            logger.output,
            [
                "DEBUG:dakara_player.web_client:"
                "Telling the server that the transition of playlist entry "
                "42 has started"
            ],
        )

        # assert the call
        mocked_put.assert_called_with(
            endpoint="playlist/player/status/",
            data={"event": "started_transition", "playlist_entry_id": 42},
            message_on_error=(
                "Unable to report that the transition of a playlist entry has started"
            ),
        )

    @patch.object(HTTPClientDakara, "put")
    def test_put_status_started_transition_failed(self, mocked_put):
        """Test to report that an invalid playlist entry transition started."""
        # call the method
        with self.assertRaises(AssertionError):
            self.http_client.put_status_started_transition(None)

        # assert the call
        mocked_put.assert_not_called()

    @patch.object(HTTPClientDakara, "put")
    def test_put_status_started_song_successful(self, mocked_put):
        """Test to report that a playlist entry song started."""
        # call the method
        with self.assertLogs("dakara_player.web_client", "DEBUG") as logger:
            self.http_client.put_status_started_song(42)

        # assert the effect on logs
        self.assertListEqual(
            logger.output,
            [
                "DEBUG:dakara_player.web_client:"
                "Telling the server that the song of playlist entry 42 has started"
            ],
        )

        # assert the call
        mocked_put.assert_called_with(
            endpoint="playlist/player/status/",
            data={"event": "started_song", "playlist_entry_id": 42},
            message_on_error=(
                "Unable to report that the song of a playlist entry has started"
            ),
        )

    @patch.object(HTTPClientDakara, "put")
    def test_put_status_started_song_failed(self, mocked_put):
        """Test to report that an invalid playlist entry song started."""
        # call the method
        with self.assertRaises(AssertionError):
            self.http_client.put_status_started_song(None)

        # assert the call
        mocked_put.assert_not_called()

    @patch.object(HTTPClientDakara, "put")
    def test_put_status_could_not_play_successful(self, mocked_put):
        """Test to report that a playlist entry could not be played."""
        # call the method
        with self.assertLogs("dakara_player.web_client", "DEBUG") as logger:
            self.http_client.put_status_could_not_play(42)

        # assert the effect on logs
        self.assertListEqual(
            logger.output,
            [
                "DEBUG:dakara_player.web_client:"
                "Telling the server that the playlist entry 42 could not play"
            ],
        )

        # assert the call
        mocked_put.assert_called_with(
            endpoint="playlist/player/status/",
            data={"event": "could_not_play", "playlist_entry_id": 42},
            message_on_error="Unable to report that playlist entry could not play",
        )

    @patch.object(HTTPClientDakara, "put")
    def test_put_status_could_not_play_failed(self, mocked_put):
        """Test to report that an invalid playlist entry could not be played."""
        # call the method
        with self.assertRaises(AssertionError):
            self.http_client.put_status_could_not_play(None)

        # assert the call
        mocked_put.assert_not_called()

    @patch.object(HTTPClientDakara, "put")
    def test_put_status_paused_successful(self, mocked_put):
        """Test to report that the player paused."""
        # call the method
        with self.assertLogs("dakara_player.web_client", "DEBUG") as logger:
            self.http_client.put_status_paused(42, 424242)

        # assert the effect on logs
        self.assertListEqual(
            logger.output,
            [
                "DEBUG:dakara_player.web_client:"
                "Telling the server that the player is paused"
            ],
        )

        # assert the call
        mocked_put.assert_called_with(
            endpoint="playlist/player/status/",
            data={"event": "paused", "playlist_entry_id": 42, "timing": 424242},
            message_on_error="Unable to report that the player is paused",
        )

    @patch.object(HTTPClientDakara, "put")
    def test_put_status_paused_failed(self, mocked_put):
        """Test to report that the player paused with incorrect entry."""
        # call the method
        with self.assertRaises(AssertionError):
            self.http_client.put_status_paused(None, 424242)

        # assert the call
        mocked_put.assert_not_called()

    @patch.object(HTTPClientDakara, "put")
    def test_put_status_resumed_successful(self, mocked_put):
        """Test to report that the player resumed playing."""
        # call the method
        with self.assertLogs("dakara_player.web_client", "DEBUG") as logger:
            self.http_client.put_status_resumed(42, 424242)

        # assert the effect on logs
        self.assertListEqual(
            logger.output,
            [
                "DEBUG:dakara_player.web_client:"
                "Telling the server that the player resumed playing"
            ],
        )

        # assert the call
        mocked_put.assert_called_with(
            endpoint="playlist/player/status/",
            data={"event": "resumed", "playlist_entry_id": 42, "timing": 424242},
            message_on_error="Unable to report that the player resumed playing",
        )

    @patch.object(HTTPClientDakara, "put")
    def test_put_status_resumed_failed(self, mocked_put):
        """Test to report that the player resumed playing with incorrect entry."""
        # call the method
        with self.assertRaises(AssertionError):
            self.http_client.put_status_resumed(None, 424242)

        # assert the call
        mocked_put.assert_not_called()

    @patch.object(HTTPClientDakara, "put")
    def test_put_status_updated_timing_successful(self, mocked_put):
        """Test to report that the player updated its timing."""
        # call the method
        with self.assertLogs("dakara_player.web_client", "DEBUG") as logger:
            self.http_client.put_status_update_timing(42, 424242)

        # assert the effect on logs
        self.assertListEqual(
            logger.output,
            [
                "DEBUG:dakara_player.web_client:"
                "Telling the server that the player updated its timing"
            ],
        )

        # assert the call
        mocked_put.assert_called_with(
            endpoint="playlist/player/status/",
            data={"event": "updated_timing", "playlist_entry_id": 42, "timing": 424242},
            message_on_error="Unable to report that the player updated its timing",
        )

    @patch.object(HTTPClientDakara, "put")
    def test_put_status_updated_timing_failed(self, mocked_put):
        """Test to report that the player updated its timing with incorrect entry."""
        # call the method
        with self.assertRaises(AssertionError):
            self.http_client.put_status_update_timing(None, 424242)

        # assert the call
        mocked_put.assert_not_called()


class WebSocketClientDakaraTestCase(TestCase):
    """Test the WebSocket client."""

    def setUp(self):
        # create a mock websocket
        self.websocket = MagicMock()

        # create a server address
        self.address = "www.example.com"

        # create an URL
        self.url = "ws://www.example.com/ws"

        # create token header
        self.header = {"token": "token"}

        # create a reconnect interval
        self.reconnect_interval = 1

        # create stop event and errors queue
        self.stop = Event()
        self.errors = Queue()

        # create a WebSocketClientDakara instance
        self.websocket_client = WebSocketClientDakara(
            self.stop,
            self.errors,
            {"address": self.address, "reconnect_interval": self.reconnect_interval},
            header=self.header,
            endpoint="ws",
        )

    def test_init_url(self):
        """Test the URL of the created object."""
        self.assertEqual(self.websocket_client.server_url, self.url)

    @patch.object(WebSocketClientDakara, "send_ready")
    def test_on_connected(self, mocked_send_ready):
        """Test the callback on connection open."""
        # call the method
        self.websocket_client.on_connected()

        # assert the call
        mocked_send_ready.assert_called_once_with()

    @patch.object(WebSocketClientDakara, "create_timer")
    def test_on_connection_lost(self, mocked_create_timer):
        """Test the callback on connection lost."""
        # mock the callback
        mocked_connection_lost = MagicMock()
        self.websocket_client.set_callback("connection_lost", mocked_connection_lost)

        # call the on_close callback
        with self.assertLogs("dakara_base.websocket_client", "DEBUG"):
            self.websocket_client.on_close(None, None)

        # assert the call
        mocked_connection_lost.assert_called_once_with()
        mocked_create_timer.assert_called_once_with(ANY, ANY)

    def test_receive_idle(self):
        """Test the receive idle event method."""
        # mock the callback
        mocked_idle_callback = MagicMock()
        self.websocket_client.set_callback("idle", mocked_idle_callback)

        # call the method
        with self.assertLogs("dakara_player.web_client", "DEBUG") as logger:
            self.websocket_client.receive_idle({})

        # assert the effect on logs
        self.assertListEqual(
            logger.output, ["DEBUG:dakara_player.web_client:Received idle order"]
        )

        # assert the call
        mocked_idle_callback.assert_called_once_with()

    def test_receive_playlist_entry(self):
        """Test the receive new playlist entry event method."""
        content = {"id": 0, "song": None}

        # mock the callback
        mocked_playlist_entry_callback = MagicMock()
        self.websocket_client.set_callback(
            "playlist_entry", mocked_playlist_entry_callback
        )

        # call the method
        with self.assertLogs("dakara_player.web_client", "DEBUG") as logger:
            self.websocket_client.receive_playlist_entry(content)

        # assert the effect on logs
        self.assertListEqual(
            logger.output,
            ["DEBUG:dakara_player.web_client:Received new playlist entry 0 order"],
        )

        # assert the call
        mocked_playlist_entry_callback.assert_called_with(content)

    def test_receive_command(self):
        """Test the receive command event method."""
        content = {"command": "command_value"}

        # mock the callback
        mocked_command_callback = MagicMock()
        self.websocket_client.set_callback("command", mocked_command_callback)

        # call the method
        with self.assertLogs("dakara_player.web_client", "DEBUG") as logger:
            self.websocket_client.receive_command(content)

        # assert the effect on logs
        self.assertListEqual(
            logger.output,
            ["DEBUG:dakara_player.web_client:Received command command_value order"],
        )

        # assert the call
        mocked_command_callback.assert_called_with("command_value")

    @patch.object(WebSocketClientDakara, "send")
    def test_send_ready(self, mocked_send):
        """Test to notify the server that the player is ready."""
        # call the command
        with self.assertLogs("dakara_player.web_client", "DEBUG") as logger:
            self.websocket_client.send_ready()

        # assert the effect on logs
        self.assertListEqual(
            logger.output,
            [
                "DEBUG:dakara_player.web_client:"
                "Telling the server that the player is ready"
            ],
        )

        # assert the call
        mocked_send.assert_called_with("ready")



"""VLC media player."""

import json
import logging
import platform
import re

from dakara_base.exceptions import DakaraError
from dakara_base.safe_workers import safe
from packaging.version import parse

from dakara_player.window import DummyWindowManager, WindowManager

try:
    import vlc
    from vlc import libvlc_get_version

except (ImportError, OSError):
    vlc = None
    libvlc_get_version = None

from dakara_player.media_player.base import (
    InvalidStateError,
    MediaPlayer,
    VersionNotFoundError,
    on_playing_this,
)
from dakara_player.mrl import mrl_to_path, path_to_mrl

try:
    METADATA_KEYS_COUNT = len(vlc.Meta.__dict__["_enum_names_"])

except AttributeError:
    METADATA_KEYS_COUNT = 0


logger = logging.getLogger(__name__)


class MediaPlayerVlc(MediaPlayer):
    """Class to manipulate VLC.

    The class can be used as a context manager that closes VLC
    automatically on exit.

    Any exception in callbacks make the application to crash.

    Args:
        stop (threading.Event): Stop event that notify to stop the entire
            program when set.
        errors (queue.Queue): Error queue to communicate the exception to the
            main thread.
        config (dict): Dictionary of configuration.
        tempdir (pathlib.Path): Path of the temporary directory.

    Attributes:
        stop (threading.Event): Stop event that notify to stop the entire
            program when set.
        errors (queue.Queue): Error queue to communicate the exception to the
            main thread.
        player_name (str): Name of VLC.
        fullscreen (bool): If `True`, VLC will be fullscreen.
        kara_folder_path (pathlib.Path): Path to the karaoke folder.
        playlist_entry (dict): Playlist entyr object.
        callbacks (dict): High level callbacks associated with the media
            player.
        warn_long_exit (bool): If `True`, display a warning message if the media
            player takes too long to stop.
        durations (dict of int): Duration of the different screens in seconds.
        text_paths (dict of pathlib.Path): Path of the different text screens.
        text_generator (dakara_player.text_generator.TextGenerator): Text
            generator instance.
        background_loader
        (dakara_player.background_loader.BackgroundLoader): Background
            loader instance.
        media_parameters (list of str): Extra parameters passed to the media.
        instance (vlc.Instance): Instance of VLC.
        player (vlc.MediaPlayer): VLC player.
        event_manager (vlc.EventManager): VLC event manager.
        vlc_callbacks (dict): Low level callbacks associated with VLC.
        playlist_entry_data (dict): Extra data of the playlist entry.
    """

    player_name = "VLC"

    @staticmethod
    def is_available():
        """Indicate if VLC is available.

        Returns:
            bool: `True` if VLC is useable.
        """
        try:
            return vlc is not None and vlc.Instance() is not None
        except NameError:
            logger.exception("Failed to start VLC.")
            return False

    def init_player(self, config, tempdir):
        """Initialize the objects of VLC.

        Actions performed in this method should not have any side effects
        (query file system, etc.).

        Args:
            config (dict): Dictionary of configuration.
            tempdir (pathlib.Path): Path of the temporary directory.
        """
        # parameters
        config_vlc = config.get("vlc") or {}
        self.media_parameters = config_vlc.get("media_parameters") or []

        # window for VLC
        if config_vlc.get("use_default_window", False):
            window_manager_class = DummyWindowManager

        else:
            window_manager_class = WindowManager

        self.window = window_manager_class(
            title="Dakara Player VLC",
            fullscreen=self.fullscreen,
        )

        # VLC objects
        self.instance = vlc.Instance(config_vlc.get("instance_parameters") or [])
        self.player = self.instance.media_player_new()
        self.event_manager = self.player.event_manager()

        # vlc callbacks
        self.vlc_callbacks = {}

        # playlist entry objects
        self.playlist_entry_data = {}
        self.clear_playlist_entry_player()

    def load_player(self):
        """Perform actions with side effects for VLC initialization."""
        # check VLC version
        self.check_version()

        # assign window to VLC
        self.window.open()
        self.set_window(self.window.get_id())

        # set VLC callbacks
        self.set_vlc_default_callbacks()

        # print VLC version
        logger.info("VLC %s", self.get_version_str())

    @on_playing_this(["song"], default_return=0)
    def get_timing(self):
        """Get VLC timing.

        Returns:
            int: Current song timing in seconds if a song is playing, or 0 when
                idle or during transition screen.
        """
        timing = self.player.get_time()

        # correct the way VLC handles when it hasn't started to play yet
        if timing == -1:
            timing = 0

        return timing // 1000

    @staticmethod
    def get_version():
        """Get VLC version.

        VLC version given by the lib is on the form "x.y.z CodeName" in bytes.

        Returns:
            packaging.version.Version: Parsed version of VLC.

        Raises:
            VersionNotFoundError: If the version cannot be parsed.
        """
        match = re.search(r"(\d+\.\d+\.\d+(?:\.\d+)*)", libvlc_get_version().decode())
        if match:
            return parse(match.group(1))

        raise VersionNotFoundError("Unable to get VLC version")

    def check_version(self):
        """Check that VLC is at least version 3.

        Raises:
            VlcTooOldError: If VLC version is lower than 3.
        """
        version = self.get_version()

        if version.major < 3:
            raise VlcTooOldError("VLC is too old (version 3 and higher supported)")

        if version.minor == 0 and 13 <= version.micro <= 16:
            logger.warning(
                "This version of VLC is known to not work with Dakara player"
            )

    def set_vlc_default_callbacks(self):
        """Set VLC default callbacks."""
        self.set_vlc_callback(
            vlc.EventType.MediaPlayerEndReached, self.handle_end_reached
        )
        self.set_vlc_callback(
            vlc.EventType.MediaPlayerEncounteredError, self.handle_encountered_error
        )
        self.set_vlc_callback(vlc.EventType.MediaPlayerPlaying, self.handle_playing)
        self.set_vlc_callback(vlc.EventType.MediaPlayerPaused, self.handle_paused)

    def set_vlc_callback(self, event, callback):
        """Assing an arbitrary callback to a VLC event.

        Callback is attached to the VLC event manager and added to the
        `vlc_callbacks` dictionary.

        Args:
            event (vlc.EventType): VLC event to attach the callback to, name of
                the callback in the `vlc_callbacks` attribute.
            callback (function): Function to assign.
        """
        self.vlc_callbacks[event] = callback
        self.event_manager.event_attach(event, callback)

    def is_playing(self):
        """Query if VLC is playing something.

        Returns:
            bool: `True` if VLC is playing something.
        """
        return self.player.get_state() == vlc.State.Playing

    def is_paused(self):
        """Query if VLC is paused.

        Returns:
            bool: True if VLC is paused.
        """
        return self.player.get_state() == vlc.State.Paused

    def is_playing_this(self, what):
        """Query if VLC is playing the requested media type.

        Args:
            what (str): Tell if VLC current track is of the requested type, but
                not if it is actually playing it (it can be in pause).

        Returns:
            bool: `True` if VLC is playing the requested type.
        """
        media = self.player.get_media()

        # if no media is in the player
        if not media:
            return False

        # if no playlist entry is supposed to play
        if what in ("transition", "song") and not self.playlist_entry_data[what].media:
            return False

        return get_metadata(media)["type"] == what

    def play(self, what):
        """Request VLC to play something.

        No preparation should be done by this function, i.e. the media track
        should have been prepared already by `set_playlist_entry_player`.

        Args:
            what (str): What media to play.

        Raises:
            ValueError: If the action to play is unknown.
        """
        if what == "idle":
            # create idle screen media
            media = self.instance.media_new_path(
                str(self.background_loader.backgrounds["idle"])
            )

            media.add_options(
                *self.media_parameters,
                "image-duration={}".format(self.durations["idle"]),
                "sub-file={}".format(self.text_paths["idle"]),
                "no-sub-autodetect-file",
            )

            set_metadata(media, {"type": "idle"})

            self.generate_text("idle")

        elif what == "transition":
            media = self.playlist_entry_data["transition"].media

        elif what == "song":
            media = self.playlist_entry_data["song"].media

        else:
            raise ValueError("Unexpected action to play: {}".format(what))

        self.player.set_media(media)
        self.player.play()

    @on_playing_this(["transition", "song"])
    def pause(self):
        """Request VLC to pause.

        Can only work on transition screens or songs. Pausing should have no
        effect if VLC is already paused.
        """
        if self.is_paused():
            logger.debug("Player already in pause")
            return

        logger.info("Setting pause")
        self.player.pause()

    @on_playing_this(["transition", "song"])
    def resume(self):
        """Request VLC to resume playing.

        Can only work on transition screens or songs. Resuming should have no
        effect if VLC is already playing.
        """
        if not self.is_paused():
            logger.debug("Player already playing")
            return

        logger.info("Resuming play")
        self.player.play()

    @on_playing_this(["song"])
    def restart(self):
        """Request to restart the current media.

        Can only work on songs.
        """
        logger.info("Restarting media")
        self.player.set_time(0)
        self.callbacks["updated_timing"](self.playlist_entry["id"], self.get_timing())

    @on_playing_this(["transition", "song"])
    def skip(self, no_callback=False):
        """Request to skip the current media.

        Can only work on transition screens or songs. VLC should continue
        playing, but media has to be considered already finished.

        Args:
            no_callback (bool): If `True`, no callback to signal the song has
                finished will be executed.
        """
        logger.info("Skipping '%s'", self.playlist_entry["song"]["title"])
        if not no_callback:
            self.callbacks["finished"](self.playlist_entry["id"])

        self.clear_playlist_entry()

    @on_playing_this(["song"])
    def rewind(self):
        """Request to rewind a few seconds the media.

        Can only work on songs. It cannot rewind before the beginning of the
        media. In that case, restart the song.
        """
        timing = int(
            self.player.get_time() - self.durations["rewind_fast_forward"] * 1000
        )

        if timing < 0:
            self.restart()
            return

        logger.info("Rewinding media")
        self.player.set_time(timing)
        self.callbacks["updated_timing"](self.playlist_entry["id"], self.get_timing())

    @on_playing_this(["song"])
    def fast_forward(self):
        """Request to fast forward a few seconds the media.

        Can only work on songs. It cannot advance passed the end of the media.
        In that case, skip the song.
        """
        timing = int(
            self.player.get_time() + self.durations["rewind_fast_forward"] * 1000
        )

        if timing > self.player.get_media().get_duration():
            self.skip()
            return

        logger.info("Fast forwarding media")
        self.player.set_time(timing)
        self.callbacks["updated_timing"](self.playlist_entry["id"], self.get_timing())

    def stop_player(self):
        """Request to stop VLC."""
        # stopping VLC
        logger.info("Stopping player")
        self.player.stop()
        logger.debug("Stopped player")

        # closing window
        self.window.close()

    def set_playlist_entry_player(self, playlist_entry, file_path, autoplay):
        """Prepare playlist entry data to be played.

        Prepare all media objects, subtitles, etc. for being played, for the
        transition screen and the song. Such data should be stored on a
        dedicated object, like `playlist_entry_data`.

        Args:
            playlist_entry (dict): Playlist entry object.
            file_path (pathlib.Path): Absolute path to the song file.
            autoplay (bool): If `True`, start to play transition screen as soon
                as possible (i.e. as soon as the transition screen media is
                ready). The song media is prepared when the transition screen
                is playing.
        """
        # create transition screen media
        media_transition = self.instance.media_new_path(
            str(self.background_loader.backgrounds["transition"])
        )

        media_transition.add_options(
            *self.media_parameters,
            "image-duration={}".format(self.durations["transition"]),
            "sub-file={}".format(self.text_paths["transition"]),
            "no-sub-autodetect-file",
        )

        set_metadata(
            media_transition,
            {"type": "transition", "playlist_entry": self.playlist_entry},
        )

        self.generate_text("transition")

        self.playlist_entry_data["transition"].media = media_transition

        # start playing transition right away if requested
        if autoplay:
            self.play("transition")

        # create song media
        media_song = self.instance.media_new_path(str(file_path))
        media_song.add_options(*self.media_parameters)
        media_song.parse()
        set_metadata(
            media_song, {"type": "song", "playlist_entry": self.playlist_entry}
        )

        self.playlist_entry_data["song"].media = media_song

        # manage instrumental
        if playlist_entry["use_instrumental"]:
            self.manage_instrumental(playlist_entry, file_path)

    def manage_instrumental(self, playlist_entry, file_path):
        """Manage the requested instrumental track.

        Instrumental track is searched first in audio files having the same
        name as the video file, then in extra audio tracks of the video file.

        Args:
            playlist_entry (dict): Playlist entry data. Must contain the key
                `use_instrumental`.
            file_path (pathlib.Path): Path of the song file.
        """
        # get instrumental file if possible
        audio_path = self.get_instrumental_file(file_path)

        # if audio file is present, request to add the file to the media
        # as a slave and register to play this extra track (which will be
        # the last audio track of the media)
        if audio_path:
            number_tracks = self.get_number_tracks(
                self.playlist_entry_data["song"].media
            )
            logger.info(
                "Requesting to play instrumental file '%s' for '%s'",
                audio_path,
                file_path,
            )
            try:
                # try to add the instrumental file
                self.playlist_entry_data["song"].media.slaves_add(
                    vlc.MediaSlaveType.audio, 4, path_to_mrl(audio_path).encode()
                )

            except NameError:
                # otherwise fallback to default
                logger.error(
                    "This version of VLC does not support slaves, cannot add "
                    "instrumental file"
                )
                return

            self.playlist_entry_data["song"].track_id_audio = number_tracks
            return

        # get audio track ids
        track_id_audio_list = self.get_track_id_audio_list(
            self.playlist_entry_data["song"].media
        )

        # if more than 1 audio track is present, register to play the 2nd one
        if len(track_id_audio_list) > 1:
            logger.info("Requesting to play instrumental track of '%s'", file_path)
            self.playlist_entry_data["song"].track_id_audio = track_id_audio_list[1]
            return

        # otherwise, fallback to register to play the first track and log it
        logger.warning(
            "Cannot find instrumental file or track for file '%s'", file_path
        )

    def clear_playlist_entry_player(self):
        """Clean playlist entry data after being played."""
        self.playlist_entry_data = {
            "transition": Media(),
            "song": MediaSong(),
        }

    @staticmethod
    def get_number_tracks(media):
        """Get number of all tracks of the media.

        Args:
            media (vlc.Media): Media to investigate.

        Returns:
            int: Number of tracks in the media.
        """
        return len(list(media.tracks_get()))

    @staticmethod
    def get_track_id_audio_list(media):
        """Get ID of audio tracks of the media.

        Args:
            media (vlc.Media): Media to investigate.

        Returns:
            list of int: ID of audio tracks in the media.
        """
        return [
            item.id for item in media.tracks_get() if item.type == vlc.TrackType.audio
        ]

    @safe
    def handle_end_reached(self, event):
        """Callback called when a media ends.

        This happens when:
            - A transition screen ends, leading to playing the actual song;
            - A song ends normally, leading to calling the callback
                `callbacks["finished"]`;
            - An idle screen ends, leading to reloop it.

        A new thread is created in any case.

        Args:
            event (vlc.EventType): VLC event object.

        Raises:
            InvalidStateError: If the context of call of this callback is
                unexpected.
        """
        logger.debug("End reached callback called")

        # the transition screen has finished, request to play the song itself
        if self.is_playing_this("transition"):
            logger.debug(
                "Will play '{}'".format(
                    mrl_to_path(self.playlist_entry_data["song"].media.get_mrl())
                )
            )
            thread = self.create_thread(target=self.play, args=("song",))
            thread.start()

            return

        # the media has finished, so clean memory and call the according callback
        if self.is_playing_this("song"):
            playlist_entry_id = self.playlist_entry["id"]
            self.clear_playlist_entry()
            self.callbacks["finished"](playlist_entry_id)

            return

        # the idle screen has finished, simply restart it
        if self.is_playing_this("idle"):
            thread = self.create_thread(target=self.play, args=("idle",))
            thread.start()

            return

        # if no state can be determined, raise an error
        raise InvalidStateError("End reached on an undeterminated state")

    @safe
    def handle_encountered_error(self, event):
        """Callback called when error occurs

        There is no way to capture error message, so only a generic error
        message is provided. Call the callbacks `callbacks["finished"]` and
        `callbacks["error"]`

        Args:
            event (vlc.EventType): VLC event object.
        """
        logger.debug("Error callback called")

        # the current song media has an error, skip the song, log the error and
        # call error callback
        if self.is_playing_this("song"):
            logger.error(
                "Unable to play '%s'", mrl_to_path(self.player.get_media().get_mrl())
            )
            self.callbacks["error"](
                self.playlist_entry["id"], "Unable to play current song"
            )
            self.skip()

            return

        # do not assess other errors

    @safe
    def handle_playing(self, event):
        """Callback called when playing has started.

        This happens when:
            - The player resumes from pause;
            - A transition screen starts;
            - A song starts, leading to set the requested audio track to play;
            - An idle screen starts.

        Args:
            event (vlc.EventType): VLC event object.

        Raises:
            InvalidStateError: If the context of call of this callback is
                unexpected.
        """
        logger.debug("Playing callback called")

        # the media or the transition is resuming from pause
        # it is pretty hard to detect this case, as we do not have a previous
        # state in memory
        # we rely on a specific flag stored in `playlist_entry_data` Media
        # objects which is set to True when the corresponding media starts
        if (
            self.is_playing_this("transition")
            and self.playlist_entry_data["transition"].started
            or self.is_playing_this("song")
            and self.playlist_entry_data["song"].started
        ):
            self.callbacks["resumed"](self.playlist_entry["id"], self.get_timing())
            logger.debug("Resumed play")

            return

        # the transition screen starts to play
        if self.is_playing_this("transition"):
            self.callbacks["started_transition"](self.playlist_entry["id"])
            self.playlist_entry_data["transition"].started = True
            logger.info(
                "Playing transition for '%s'", self.playlist_entry["song"]["title"]
            )

            return

        # the song starts to play
        if self.is_playing_this("song"):
            self.callbacks["started_song"](self.playlist_entry["id"])

            # set instrumental track if necessary
            track_id_audio = self.playlist_entry_data["song"].track_id_audio
            if track_id_audio is not None:
                logger.debug("Requesting to play audio track %i", track_id_audio)
                self.player.audio_set_track(track_id_audio)

            self.playlist_entry_data["song"].started = True
            logger.info(
                "Now playing '%s' ('%s')",
                self.playlist_entry["song"]["title"],
                mrl_to_path(self.player.get_media().get_mrl()),
            )

            return

        # the idle screen starts to play
        if self.is_playing_this("idle"):
            logger.debug("Playing idle screen")

            return

        raise InvalidStateError("Playing on an undeterminated state")

    @safe
    def handle_paused(self, event):
        """Callback called when pause is set.

        Args:
            event (vlc.EventType): VLC event object.
        """
        logger.debug("Paused callback called")

        # call paused callback
        self.callbacks["paused"](self.playlist_entry["id"], self.get_timing())

        logger.debug("Paused")

    def set_window(self, id):
        """Associate an existing window to VLC

        Args:
            id (int): ID of the window.

        Raises:
            NotImplementedError: If the current platform is not supported.
        """
        if id is None:
            logger.debug("Using VLC default window")
            return

        system = platform.system()

        if system == "Linux":
            logger.debug("Associating X window to VLC")
            self.player.set_xwindow(id)
            return

        if system == "Windows":
            logger.debug("Associating Win API window to VLC")
            self.player.set_hwnd(id)
            return

        raise NotImplementedError(
            "This operating system ({}) is not currently supported".format(system)
        )


def set_metadata(media, metadata):
    """Set metadata to media.

    Take the first free metadata slot to store value. The metadata can be
    extracted after.

    Args:
        media (vlc.Media): Media to set metadata in.
        metadata (any): JSON representable data.

    Raises:
        ValueError: If the media has no free metadata to use.
    """
    for key in range(METADATA_KEYS_COUNT):
        if media.get_meta(key) is None:
            media.set_meta(key, json.dumps(metadata))
            return

    raise ValueError("This media has no spare metadata to use")


def get_metadata(media):
    """Get metadata from media.

    Take the first non free metadata slot that contains a valid JSON value.

    Args:
        media (vlc.Media): Media to get metadata from.

    Returns:
        any: JSON representable data.

    Raises:
        ValueError: If the media has no valid metadata.
    """
    for key in range(METADATA_KEYS_COUNT):
        try:
            return json.loads(media.get_meta(key))

        except (json.JSONDecodeError, TypeError):
            continue

    raise ValueError("This media has no set metadata")


class Media:
    """Media object."""

    def __init__(self, media=None):
        self.media = media
        self.started = False


class MediaSong(Media):
    """Song object."""

    def __init__(self, *args, track_id_audio=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.track_id_audio = track_id_audio


class VlcTooOldError(DakaraError):
    """Error raised if VLC is too old."""



from argparse import ArgumentParser, Namespace
from unittest import TestCase
from unittest.mock import ANY, MagicMock, patch

from dakara_base.config import Config
from dakara_base.exceptions import DakaraHandledError
from dakara_base.http_client import ParameterError

from dakara_player.__main__ import (
    create_config,
    create_resources,
    get_parser,
    main,
    play,
)


class GetParserTestCase(TestCase):
    """Test the parser creator."""

    def test(self):
        """Test a parser is created."""
        parser = get_parser()
        self.assertIsNotNone(parser)

    def test_play_function(self):
        """Test the parser calls play when prompted."""
        # call the function
        parser = get_parser()
        args = parser.parse_args(["play"])

        # check the function
        self.assertIs(args.function, play)

    def test_create_config_function(self):
        """Test the parser calls create_config when prompted."""
        # call the function
        parser = get_parser()
        args = parser.parse_args(["create-config"])

        # check the function
        self.assertIs(args.function, create_config)

    def test_create_resources_function(self):
        """Test the parser calls create_resources when prompted."""
        # call the function
        parser = get_parser()
        args = parser.parse_args(["create-resources"])

        # check the function
        self.assertIs(args.function, create_resources)


@patch("dakara_player.__main__.DakaraPlayer", autospec=True)
@patch("dakara_player.__main__.set_loglevel")
@patch.object(Config, "set_debug")
@patch.object(Config, "check_mandatory_keys")
@patch.object(Config, "load_file")
@patch("dakara_player.__main__.create_logger")
class PlayTestCase(TestCase):
    """Test the play action."""

    def test_play(
        self,
        mocked_create_logger,
        mocked_load_file,
        mocked_check_mandatory_keys,
        mocked_set_debug,
        mocked_set_loglevel,
        mocked_dakara_player_class,
    ):
        """Test the play command."""
        # call the function
        play(Namespace(debug=False))

        # assert the call
        mocked_create_logger.assert_called_with()
        mocked_load_file.assert_called_with(ANY)
        mocked_check_mandatory_keys.assert_called_with(["player", "server"])
        mocked_set_loglevel.assert_called_with(ANY)
        mocked_dakara_player_class.assert_called_with(ANY)
        mocked_dakara_player_class.return_value.load.assert_called_with()
        mocked_dakara_player_class.return_value.run.assert_called_with()

    def test_play_parameter_error(
        self,
        mocked_create_logger,
        mocked_load_file,
        mocked_check_mandatory_keys,
        mocked_set_debug,
        mocked_set_loglevel,
        mocked_dakara_player_class,
    ):
        """Test an error in HTTP server configuration."""
        # raise exception
        mocked_dakara_player_class.return_value.run.side_effect = ParameterError(
            "error message"
        )

        # call the function
        with self.assertRaisesRegex(ParameterError, "error message") as cm:
            play(Namespace(debug=False))

        self.assertIsInstance(cm.exception, ParameterError)
        self.assertIsInstance(cm.exception, DakaraHandledError)


@patch("dakara_player.__main__.create_logger")
@patch("dakara_player.__main__.create_config_file")
class CreateConfigTestCase(TestCase):
    """Test the create-config action."""

    def test_create_config(self, mocked_create_config_file, mocked_create_logger):
        """Test a simple create-config action."""
        # call the function
        with self.assertLogs("dakara_player.__main__") as logger:
            create_config(Namespace(force=False))

        # assert the logs
        self.assertListEqual(
            logger.output,
            ["INFO:dakara_player.__main__:Please edit this file"],
        )

        # assert the call
        mocked_create_logger.assert_called_with(
            custom_log_format=ANY, custom_log_level=ANY
        )
        mocked_create_config_file.assert_called_with(
            "dakara_player.resources", "player.yaml", False
        )


@patch("dakara_player.__main__.create_logger")
@patch("dakara_player.__main__.create_resource_files")
class CreateResourcesTestCase(TestCase):
    """Test the create-resources action."""

    def test_create_config(self, mocked_create_resource_files, mocked_create_logger):
        """Test a simple create-resources action."""
        # call the function
        with self.assertLogs("dakara_player.__main__") as logger:
            create_resources(Namespace(force=False))

        # assert the logs
        self.assertListEqual(
            logger.output,
            ["INFO:dakara_player.__main__:You can now customize those files"],
        )

        # assert the call
        mocked_create_logger.assert_called_with(
            custom_log_format=ANY, custom_log_level=ANY
        )
        mocked_create_resource_files.assert_called_with(False)


@patch("dakara_player.__main__.sys.exit")
@patch.object(ArgumentParser, "parse_args")
class MainTestCase(TestCase):
    """Test the main action."""

    def test_normal_exit(self, mocked_parse_args, mocked_exit):
        """Test a normal exit."""
        # create mocks
        function = MagicMock()
        mocked_parse_args.return_value = Namespace(function=function, debug=False)

        # call the function
        main()

        # assert the call
        function.assert_called_with(ANY)
        mocked_exit.assert_called_with(0)



from contextlib import ExitStack, contextmanager
from pathlib import Path
from queue import Queue
from tempfile import TemporaryDirectory
from threading import Event
from unittest import skipIf, skipUnless
from unittest.mock import MagicMock

try:
    import vlc

except (ImportError, OSError):
    vlc = None

from dakara_base.config import Config
from func_timeout import func_set_timeout

from dakara_player.media_player.base import IDLE_BG_NAME, TRANSITION_BG_NAME
from dakara_player.media_player.vlc import METADATA_KEYS_COUNT, MediaPlayerVlc
from dakara_player.mrl import mrl_to_path
from tests.integration.base import TestCasePollerKara

REWIND_FAST_FORWARD_DURATION = 0.5
REWIND_FAST_FORWARD_DELTA = 0.5
DEFAULT_DELTA = 0.2


@skipUnless(MediaPlayerVlc.is_available(), "VLC not installed")
class MediaPlayerVlcIntegrationTestCase(TestCasePollerKara):
    """Test the VLC player class in real conditions."""

    TIMEOUT = 120

    def setUp(self):
        super().setUp()

        # create instance parameter
        self.instance_parameters = [
            "--vout=vdummy",
            "--aout=adummy",
            "--text-renderer=tdummy",
        ]

        # use default window
        self.use_default_window = True

        # create fullscreen flag
        self.fullscreen = True

        # create media parameter
        self.media_parameters = []

        # create transition duration
        self.transition_duration = 1

    @contextmanager
    def get_instance(self, config=None, check_error=True):
        """Get an instance of MediaPlayerVlc.

        This method is a context manager that automatically stops the player on
        exit.

        Args:
            config (dict): Extra configuration passed to the constructor.
            check_error (bool): If true, check if the player stop event is not
                set and the error queue is empty at the end.

        Yields:
            tuple: Containing the following elements:
                MediaPlayerVlc: Instance;
                pathlib.Path: Path of the temporary directory;
                unittest.case._LoggingWatcher: Captured output.
        """

        config_full = {
            "kara_folder": str(self.kara_folder_path),
            "fullscreen": self.fullscreen,
            "vlc": {
                "instance_parameters": self.instance_parameters,
                "media_parameters": self.media_parameters,
                "use_default_window": self.use_default_window,
            },
        }

        if config:
            config_full.update(config)

        with ExitStack() as stack:
            temp = Path(stack.enter_context(TemporaryDirectory())).resolve()
            vlc_player = stack.enter_context(
                MediaPlayerVlc(
                    Event(),
                    Queue(),
                    Config("DAKARA_PLAYER", config_full),
                    temp,
                    warn_long_exit=False,
                )
            )
            output = stack.enter_context(
                self.assertLogs("dakara_player.media_player.vlc", "DEBUG")
            )
            vlc_player.load()

            yield vlc_player, temp, output

            if check_error:
                # display errors in queue if any
                if not vlc_player.errors.empty():
                    _, error, traceback = vlc_player.errors.get(5)
                    error.with_traceback(traceback)
                    raise error

                # assert no errors to fail test if any
                self.assertFalse(vlc_player.stop.is_set())

    def test_metadata_keys_count(self):
        """Test the number of metadata keys."""
        self.assertNotEqual(METADATA_KEYS_COUNT, 0)

    @func_set_timeout(TIMEOUT)
    def test_start(self):
        """Test the initial state of the player without instructions."""
        with self.get_instance() as (vlc_player, _, _):
            self.assertIsNone(vlc_player.playlist_entry)
            self.assertFalse(vlc_player.is_playing_this("idle"))
            self.assertFalse(vlc_player.is_playing_this("transition"))
            self.assertFalse(vlc_player.is_playing_this("song"))

    @func_set_timeout(TIMEOUT)
    def test_play_idle(self):
        """Test to display the idle screen."""
        with self.get_instance() as (vlc_player, temp, _):
            # pre assertions
            self.assertIsNone(vlc_player.player.get_media())
            self.assertEqual(vlc_player.player.get_state(), vlc.State.NothingSpecial)

            # call the method
            vlc_player.play("idle")

            # wait for the idle screen to start
            self.wait_is_playing(vlc_player, "idle")

            # post assertions
            self.assertEqual(vlc_player.player.get_state(), vlc.State.Playing)

            self.assertIsNotNone(vlc_player.player.get_media())
            media = vlc_player.player.get_media()
            file_path = mrl_to_path(media.get_mrl())
            self.assertEqual(file_path, temp / IDLE_BG_NAME)

            # TODO check which subtitle file is read
            # seems impossible to do for now

    @func_set_timeout(TIMEOUT)
    def test_play_playlist_entry(self):
        """Test to play a playlist entry.

        First, the transition screen is played, then the song itself.
        """
        with self.get_instance() as (vlc_player, temp, _):
            # mock the callbacks
            vlc_player.set_callback("started_transition", MagicMock())
            vlc_player.set_callback("started_song", MagicMock())

            # pre assertions
            self.assertIsNone(vlc_player.playlist_entry)
            self.assertIsNone(vlc_player.player.get_media())
            self.assertEqual(vlc_player.player.get_state(), vlc.State.NothingSpecial)

            # call the method
            vlc_player.set_playlist_entry(self.playlist_entry1, autoplay=False)

            # check media did not started
            self.assertFalse(vlc_player.playlist_entry_data["transition"].started)
            self.assertFalse(vlc_player.playlist_entry_data["song"].started)

            # start playing
            vlc_player.play("transition")

            # wait for the transition screen to start
            self.wait_is_playing(vlc_player, "transition")

            # post assertions for transition screen
            self.assertEqual(vlc_player.player.get_state(), vlc.State.Playing)
            self.assertIsNotNone(vlc_player.playlist_entry)

            # check transition media only started
            self.assertTrue(vlc_player.playlist_entry_data["transition"].started)
            self.assertFalse(vlc_player.playlist_entry_data["song"].started)

            # check media exists
            media = vlc_player.player.get_media()
            self.assertIsNotNone(media)

            # check media path
            file_path = mrl_to_path(media.get_mrl())
            self.assertEqual(file_path, temp / TRANSITION_BG_NAME)

            # check there is no audio track
            track = vlc_player.player.audio_get_track()
            self.assertEqual(track, -1)

            # TODO check which subtitle file is read
            # seems impossible to do for now

            # assert the started transition callback has been called
            vlc_player.callbacks["started_transition"].assert_called_with(
                self.playlist_entry1["id"]
            )

            # wait for the media to start
            self.wait_is_playing(vlc_player, "song")

            # post assertions for song
            self.assertEqual(vlc_player.player.get_state(), vlc.State.Playing)

            # check song media also started
            self.assertTrue(vlc_player.playlist_entry_data["transition"].started)
            self.assertTrue(vlc_player.playlist_entry_data["song"].started)

            # check media exists
            media = vlc_player.player.get_media()
            self.assertIsNotNone(media)

            # check media path
            file_path = mrl_to_path(media.get_mrl())
            self.assertEqual(file_path, self.song1_path)

            # check audio track
            track = vlc_player.player.audio_get_track()
            self.assertEqual(track, 1)

            # assert the started song callback has been called
            vlc_player.callbacks["started_song"].assert_called_with(
                self.playlist_entry1["id"]
            )

            # assert the player is playing a song
            self.assertFalse(vlc_player.is_playing_this("transition"))
            self.assertTrue(vlc_player.is_playing_this("song"))
            self.assertFalse(vlc_player.is_playing_this("idle"))

            # wait for the media to end
            self.wait(vlc_player, lambda: not vlc_player.is_playing())

            # assert the player is not playing anything
            self.assertFalse(vlc_player.is_playing_this("transition"))
            self.assertFalse(vlc_player.is_playing_this("song"))
            self.assertFalse(vlc_player.is_playing_this("idle"))

    @func_set_timeout(TIMEOUT)
    def test_play_playlist_entry_instrumental_track(self):
        """Test to play a playlist entry using instrumental track."""
        # request to use instrumental track
        self.playlist_entry1["use_instrumental"] = True

        with self.get_instance() as (vlc_player, _, _):
            # mock the callbacks
            vlc_player.set_callback("started_transition", MagicMock())
            vlc_player.set_callback("started_song", MagicMock())

            # pre assertions
            self.assertIsNone(vlc_player.playlist_entry)
            self.assertIsNone(vlc_player.player.get_media())
            self.assertEqual(vlc_player.player.get_state(), vlc.State.NothingSpecial)

            # call the method
            vlc_player.set_playlist_entry(self.playlist_entry1)

            # wait for the song to start
            self.wait_is_playing(vlc_player, "song")

            # post assertions for song
            self.assertEqual(vlc_player.player.get_state(), vlc.State.Playing)

            # check media exists
            media = vlc_player.player.get_media()
            self.assertIsNotNone(media)

            # check media path
            file_path = mrl_to_path(media.get_mrl())
            self.assertEqual(file_path, self.song1_path)

            # check audio track
            track = vlc_player.player.audio_get_track()
            self.assertEqual(track, 2)

            # assert the started song callback has been called
            vlc_player.callbacks["started_song"].assert_called_with(
                self.playlist_entry1["id"]
            )

    @func_set_timeout(TIMEOUT)
    def test_play_playlist_entry_instrumental_track_avi(self):
        """Test to play a playlist entry AVI file using instrumental track.

        This type of file is known to have data in the key we use to store
        information in media.
        """
        # request to use instrumental track
        self.playlist_entry3["use_instrumental"] = True

        with self.get_instance() as (vlc_player, _, _):
            # mock the callbacks
            vlc_player.set_callback("started_transition", MagicMock())
            vlc_player.set_callback("started_song", MagicMock())

            # pre assertions
            self.assertIsNone(vlc_player.playlist_entry)
            self.assertIsNone(vlc_player.player.get_media())
            self.assertEqual(vlc_player.player.get_state(), vlc.State.NothingSpecial)

            # call the method
            vlc_player.set_playlist_entry(self.playlist_entry3)

            # wait for the song to start
            self.wait_is_playing(vlc_player, "song")

            # post assertions for song
            self.assertEqual(vlc_player.player.get_state(), vlc.State.Playing)

            # check media exists
            media = vlc_player.player.get_media()
            self.assertIsNotNone(media)

            # check media path
            file_path = mrl_to_path(media.get_mrl())
            self.assertEqual(file_path, self.song3_path)

            # check audio track
            track = vlc_player.player.audio_get_track()
            self.assertEqual(track, 2)

            # assert the started song callback has been called
            vlc_player.callbacks["started_song"].assert_called_with(
                self.playlist_entry3["id"]
            )

    @skipIf(
        not hasattr(vlc, "libvlc_media_slaves_add"), "VLC does not support slaves_add"
    )
    @func_set_timeout(TIMEOUT)
    def test_play_playlist_entry_instrumental_file(self):
        """Test to play a playlist entry using instrumental file."""
        # request to use instrumental file
        self.playlist_entry1["song"]["file_path"] = str(self.song2_path)
        self.playlist_entry1["use_instrumental"] = True

        with self.get_instance() as (vlc_player, _, _):
            # mock the callbacks
            vlc_player.set_callback("started_transition", MagicMock())
            vlc_player.set_callback("started_song", MagicMock())

            # pre assertions
            self.assertIsNone(vlc_player.playlist_entry)
            self.assertIsNone(vlc_player.player.get_media())
            self.assertEqual(vlc_player.player.get_state(), vlc.State.NothingSpecial)

            # call the method
            vlc_player.set_playlist_entry(self.playlist_entry1)

            # wait for the song to start
            self.wait_is_playing(vlc_player, "song")

            # post assertions for song
            self.assertEqual(vlc_player.player.get_state(), vlc.State.Playing)

            # check media exists
            media = vlc_player.player.get_media()
            self.assertIsNotNone(media)

            # check media path
            file_path = mrl_to_path(media.get_mrl())
            self.assertEqual(file_path, self.song2_path)

            # check audio track
            track = vlc_player.player.audio_get_track()
            self.assertEqual(track, 4)

            # assert the started song callback has been called
            vlc_player.callbacks["started_song"].assert_called_with(
                self.playlist_entry1["id"]
            )

    @func_set_timeout(TIMEOUT)
    def test_pause(self):
        """Test to pause and resume the player."""
        with self.get_instance() as (vlc_player, _, _):
            # mock the callbacks
            vlc_player.set_callback("paused", MagicMock())
            vlc_player.set_callback("resumed", MagicMock())

            # start the playlist entry
            vlc_player.set_playlist_entry(self.playlist_entry1)

            # wait for the song to start
            self.wait_is_playing(vlc_player, "song")

            # call the method to pause the player
            vlc_player.pause()
            timing = vlc_player.get_timing()

            # wait for the player to be paused
            self.wait_is_paused(vlc_player)

            # assert in pause
            self.assertFalse(vlc_player.is_playing())

            # assert the callback
            vlc_player.callbacks["paused"].assert_called_with(
                self.playlist_entry1["id"], timing
            )
            vlc_player.callbacks["resumed"].assert_not_called()

            # reset the mocks
            vlc_player.callbacks["paused"].reset_mock()
            vlc_player.callbacks["resumed"].reset_mock()

            # call the method to resume the player
            vlc_player.resume()

            # wait for the player to play again
            self.wait_is_playing(vlc_player)

            # assert the callback
            vlc_player.callbacks["paused"].assert_not_called()
            vlc_player.callbacks["resumed"].assert_called_with(
                self.playlist_entry1["id"],
                timing,  # on a slow computer, the timing may be inaccurate
            )

    @func_set_timeout(TIMEOUT)
    def test_double_pause(self):
        """Test that double pause and double resume have no effects."""
        with self.get_instance() as (vlc_player, _, _):
            # mock the callbacks
            vlc_player.set_callback("paused", MagicMock())
            vlc_player.set_callback("resumed", MagicMock())

            # start the playlist entry
            vlc_player.set_playlist_entry(self.playlist_entry1)

            # wait for the song to start
            self.wait_is_playing(vlc_player, "song")

            # call the method to pause the player
            vlc_player.pause()

            # wait for the player to be paused
            self.wait_is_paused(vlc_player)

            # assert the callback
            vlc_player.callbacks["paused"].assert_called()
            vlc_player.callbacks["resumed"].assert_not_called()

            # reset the mocks
            vlc_player.callbacks["paused"].reset_mock()
            vlc_player.callbacks["resumed"].reset_mock()

            # re-call the method to pause the player
            vlc_player.pause()

            # wait again for the player to be paused
            self.wait_is_paused(vlc_player)

            # assert the callback
            vlc_player.callbacks["paused"].assert_not_called()
            vlc_player.callbacks["resumed"].assert_not_called()

            # reset the mocks
            vlc_player.callbacks["paused"].reset_mock()
            vlc_player.callbacks["resumed"].reset_mock()

            # call the method to resume the player
            vlc_player.resume()

            # wait for the player to play again
            self.wait_is_playing(vlc_player)

            # assert the callback
            vlc_player.callbacks["paused"].assert_not_called()
            vlc_player.callbacks["resumed"].assert_called()

            # reset the mocks
            vlc_player.callbacks["paused"].reset_mock()
            vlc_player.callbacks["resumed"].reset_mock()

            # re-call the method to resume the player
            vlc_player.resume()
            self.wait_is_playing(vlc_player)

            # assert the callback
            vlc_player.callbacks["paused"].assert_not_called()
            vlc_player.callbacks["resumed"].assert_not_called()

    @func_set_timeout(TIMEOUT)
    def test_restart_song(self):
        """Test to restart a playlist entry."""
        with self.get_instance() as (vlc_player, _, _):
            # mock the callbacks
            vlc_player.set_callback("started_transition", MagicMock())
            vlc_player.set_callback("started_song", MagicMock())
            vlc_player.set_callback("finished", MagicMock())
            vlc_player.set_callback("updated_timing", MagicMock())

            # pre assertions
            self.assertIsNone(vlc_player.playlist_entry)
            self.assertIsNone(vlc_player.player.get_media())
            self.assertEqual(vlc_player.player.get_state(), vlc.State.NothingSpecial)

            # request playlist entry to play
            vlc_player.set_playlist_entry(self.playlist_entry1)

            # wait for the media to start
            self.wait_is_playing(vlc_player, "song")

            # post assertions for song
            self.assertEqual(vlc_player.player.get_state(), vlc.State.Playing)
            self.assertIsNotNone(vlc_player.playlist_entry)

            # wait a bit for the player to play
            self.wait(
                vlc_player,
                lambda: vlc_player.player.get_time()
                >= REWIND_FAST_FORWARD_DURATION * 1000,
            )

            # request to restart media
            vlc_player.restart()

            # check timing is 0
            self.assertAlmostEqual(vlc_player.player.get_time(), 0, delta=DEFAULT_DELTA)

            # check the song is not stopped
            self.assertIsNotNone(vlc_player.playlist_entry)
            vlc_player.callbacks["finished"].assert_not_called()

            # assert callback
            vlc_player.callbacks["updated_timing"].assert_called_with(
                self.playlist_entry1["id"], 0
            )

    @func_set_timeout(TIMEOUT)
    def test_skip_song(self):
        """Test to skip a playlist entry."""
        with self.get_instance() as (vlc_player, _, _):
            # mock the callbacks
            vlc_player.set_callback("started_transition", MagicMock())
            vlc_player.set_callback("started_song", MagicMock())
            vlc_player.set_callback("finished", MagicMock())

            # pre assertions
            self.assertIsNone(vlc_player.playlist_entry)
            self.assertIsNone(vlc_player.player.get_media())
            self.assertEqual(vlc_player.player.get_state(), vlc.State.NothingSpecial)

            # request initial playlist entry to play
            vlc_player.set_playlist_entry(self.playlist_entry1)

            # wait for the media to start
            self.wait_is_playing(vlc_player, "song")

            # post assertions for song
            self.assertEqual(vlc_player.player.get_state(), vlc.State.Playing)
            self.assertIsNotNone(vlc_player.playlist_entry)

            # check media exists
            media = vlc_player.player.get_media()
            self.assertIsNotNone(media)

            # check media path
            file_path = mrl_to_path(media.get_mrl())
            self.assertEqual(file_path, self.song1_path)

            # request first playlist entry to stop
            vlc_player.skip()

            # check the song is stopped accordingly
            self.assertIsNone(vlc_player.playlist_entry)
            vlc_player.callbacks["finished"].assert_called_with(
                self.playlist_entry1["id"]
            )

            # request second playlist entry to play
            vlc_player.set_playlist_entry(self.playlist_entry2)

            # wait for the media to start
            self.wait_is_playing(vlc_player, "song")

            # post assertions for song
            self.assertEqual(vlc_player.player.get_state(), vlc.State.Playing)
            self.assertIsNotNone(vlc_player.playlist_entry)

            # check media exists
            media = vlc_player.player.get_media()
            self.assertIsNotNone(media)

            # check media path
            file_path = mrl_to_path(media.get_mrl())
            self.assertEqual(file_path, self.song2_path)

    @func_set_timeout(TIMEOUT)
    def test_skip_transition(self):
        """Test to skip a playlist entry transition screen."""
        with self.get_instance() as (vlc_player, _, _):
            # mock the callbacks
            vlc_player.set_callback("started_transition", MagicMock())
            vlc_player.set_callback("started_song", MagicMock())
            vlc_player.set_callback("finished", MagicMock())

            # pre assertions
            self.assertIsNone(vlc_player.playlist_entry)
            self.assertIsNone(vlc_player.player.get_media())
            self.assertEqual(vlc_player.player.get_state(), vlc.State.NothingSpecial)

            # request initial playlist entry to play
            vlc_player.set_playlist_entry(self.playlist_entry1)

            # wait for the transition to start
            self.wait_is_playing(vlc_player, "transition")

            # request first playlist entry to stop
            vlc_player.skip()

            # check the song is stopped accordingly
            vlc_player.callbacks["finished"].assert_called_with(
                self.playlist_entry1["id"]
            )

            # request second playlist entry to play
            vlc_player.set_playlist_entry(self.playlist_entry2)

            # wait for the media to start
            self.wait_is_playing(vlc_player, "song")

            # post assertions for song
            self.assertEqual(vlc_player.player.get_state(), vlc.State.Playing)

            # check media exists
            media = vlc_player.player.get_media()
            self.assertIsNotNone(media)

            # check media path
            file_path = mrl_to_path(media.get_mrl())
            self.assertEqual(file_path, self.song2_path)

    @func_set_timeout(TIMEOUT)
    def test_rewind_song(self):
        """Test to rewind a playlist entry."""
        with self.get_instance(
            {
                "durations": {
                    "rewind_fast_forward_duration": REWIND_FAST_FORWARD_DURATION
                }
            }
        ) as (vlc_player, _, _):
            # mock the callbacks
            vlc_player.set_callback("started_transition", MagicMock())
            vlc_player.set_callback("started_song", MagicMock())
            vlc_player.set_callback("updated_timing", MagicMock())

            # pre assertions
            self.assertIsNone(vlc_player.playlist_entry)
            self.assertIsNone(vlc_player.player.get_media())
            self.assertEqual(vlc_player.player.get_state(), vlc.State.NothingSpecial)

            # request playlist entry to play
            vlc_player.set_playlist_entry(self.playlist_entry1)

            # wait for the media to start
            self.wait_is_playing(vlc_player, "song")

            # post assertions for song
            self.assertEqual(vlc_player.player.get_state(), vlc.State.Playing)
            self.assertIsNotNone(vlc_player.playlist_entry)

            # wait a bit for the player to play
            self.wait(
                vlc_player,
                lambda: vlc_player.player.get_time()
                >= REWIND_FAST_FORWARD_DURATION * 2 * 1000,
            )
            timing1 = vlc_player.player.get_time() / 1000

            # request playlist entry to rewind
            vlc_player.rewind()

            # check timing is earlier than previously
            timing2 = vlc_player.player.get_time() / 1000
            self.assertLess(timing2, timing1)
            self.assertAlmostEqual(
                timing1 - timing2,
                REWIND_FAST_FORWARD_DURATION,
                delta=REWIND_FAST_FORWARD_DELTA,
            )

    @func_set_timeout(TIMEOUT)
    def test_rewind_song_before_start(self):
        """Test to rewind a playlist entry before its start."""
        with self.get_instance() as (vlc_player, _, _):
            # mock the callbacks
            vlc_player.set_callback("started_transition", MagicMock())
            vlc_player.set_callback("started_song", MagicMock())

            # pre assertions
            self.assertIsNone(vlc_player.playlist_entry)
            self.assertIsNone(vlc_player.player.get_media())
            self.assertEqual(vlc_player.player.get_state(), vlc.State.NothingSpecial)

            # request playlist entry to play
            vlc_player.set_playlist_entry(self.playlist_entry1)

            # wait for the media to start
            self.wait_is_playing(vlc_player, "song")

            # post assertions for song
            self.assertEqual(vlc_player.player.get_state(), vlc.State.Playing)
            self.assertIsNotNone(vlc_player.playlist_entry)

            # request playlist entry to rewind
            vlc_player.rewind()

            # check timing is 0
            self.assertAlmostEqual(vlc_player.player.get_time(), 0, delta=DEFAULT_DELTA)

    @func_set_timeout(TIMEOUT)
    def test_fast_forward_song(self):
        """Test to advance a playlist entry."""
        with self.get_instance(
            {
                "durations": {
                    "rewind_fast_forward_duration": REWIND_FAST_FORWARD_DURATION
                }
            }
        ) as (vlc_player, _, _):
            # mock the callbacks
            vlc_player.set_callback("started_transition", MagicMock())
            vlc_player.set_callback("started_song", MagicMock())
            vlc_player.set_callback("updated_timing", MagicMock())

            # pre assertions
            self.assertIsNone(vlc_player.playlist_entry)
            self.assertIsNone(vlc_player.player.get_media())
            self.assertEqual(vlc_player.player.get_state(), vlc.State.NothingSpecial)

            # request playlist entry to play
            vlc_player.set_playlist_entry(self.playlist_entry1)

            # wait for the media to start
            self.wait_is_playing(vlc_player, "song")

            # post assertions for song
            self.assertEqual(vlc_player.player.get_state(), vlc.State.Playing)
            self.assertIsNotNone(vlc_player.playlist_entry)

            # wait a bit for the player to play
            self.wait(
                vlc_player,
                lambda: vlc_player.player.get_time()
                >= REWIND_FAST_FORWARD_DURATION * 2 * 1000,
            )
            timing1 = vlc_player.player.get_time() / 1000

            # request playlist entry to advance
            vlc_player.fast_forward()

            # check timing is earlier than previously
            timing2 = vlc_player.player.get_time() / 1000
            self.assertGreater(timing2, timing1)
            self.assertAlmostEqual(
                timing2 - timing1,
                REWIND_FAST_FORWARD_DURATION,
                delta=REWIND_FAST_FORWARD_DELTA,
            )

            # assert callback
            vlc_player.callbacks["updated_timing"].assert_called_with(
                self.playlist_entry1["id"], int(timing2)
            )

    @func_set_timeout(TIMEOUT)
    def test_fast_forward_song_after_end(self):
        """Test to advance a playlist entry after its end."""
        with self.get_instance() as (vlc_player, _, _):
            # mock the callbacks
            vlc_player.set_callback("started_transition", MagicMock())
            vlc_player.set_callback("started_song", MagicMock())
            vlc_player.set_callback("finished", MagicMock())

            # pre assertions
            self.assertIsNone(vlc_player.playlist_entry)
            self.assertIsNone(vlc_player.player.get_media())
            self.assertEqual(vlc_player.player.get_state(), vlc.State.NothingSpecial)

            # request playlist entry to play
            vlc_player.set_playlist_entry(self.playlist_entry1)

            # wait for the media to start
            self.wait_is_playing(vlc_player, "song")

            # post assertions for song
            self.assertEqual(vlc_player.player.get_state(), vlc.State.Playing)
            self.assertIsNotNone(vlc_player.playlist_entry)

            # request playlist entry to advance
            vlc_player.fast_forward()

            # check the song has finished
            vlc_player.callbacks["finished"].assert_called_with(
                self.playlist_entry1["id"]
            )
